#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
#
"""
Some implementation notes:

- glTF defines +Y as up.
- glTF uses a right-handed coordinate system, that is, the cross product of +X and +Y yields +Z.
- The front of a glTF asset faces +Z.

- Rotations are given as quaternions stored as a tuple (x,y,z,w),
  where the w-component is the cosine of half of the rotation angle.
  For example, the quaternion [ 0.259, 0.0, 0.0, 0.966 ] describes a rotation
  about 30 degrees, around the x-axis.
- The identity rotation is (0, 0, 0, 1.0)
"""

"""
The glTF file format is supported by 
[COMSOL](https://www.comsol.com/blogs/how-to-export-and-share-your-3d-result-plots-as-gltf-files)
"""
import itertools

import numpy as np
import pygltflib
from scipy.spatial.transform import Rotation

import veux
from .canvas import Canvas, Line, Mesh, Node
from veux import utility
from veux.config import NodeStyle, MeshStyle, LineStyle, DrawStyle

GLTF_T = {
    "float32": pygltflib.FLOAT,
    "uint8":   pygltflib.UNSIGNED_BYTE,
    "uint16":  pygltflib.UNSIGNED_SHORT,
    "uint32":  pygltflib.UNSIGNED_INT,
}

EYE3 = np.eye(3, dtype="float32")


def _append_index(lst, item):
    lst.append(item)
    return len(lst) - 1

class GltfLibCanvas(Canvas):
    vertical = 2

    def __init__(self, config=None):
        self.config = config
        self._data = {}

        #                 x, y, z, scalar
        self._rotation = [0, 0, 0, 1] #[-0.7071068, 0, 0, 0.7071068]
        # equivalent rotation matrix:
        self._rotation_matrix = np.eye(3) #np.array([[1,  0, 0],
                                          #[0,  0, 1],
                                          #[0, -1, 0]])

        self.index_t = "uint32"
        self.float_t = "float32"

        self.gltf = pygltflib.GLTF2(
            scene=0,
            scenes=[pygltflib.Scene(nodes=[])],
            nodes=[],
            meshes=[],
            accessors=[],
            materials=[
                pygltflib.Material(
                    name="black",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 0, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="white",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 1, 1, 1]
                    )
                ),
                pygltflib.Material(
                    name="red",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 0, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="green",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 1, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="blue",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 0, 1, 1]
                    )
                ),
                pygltflib.Material(
                    name="gray",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0.80, 0.80, 0.80, 1]
                    )
                ),
                pygltflib.Material(
                    name="hidden",
                    alphaMode=pygltflib.BLEND,  # Enables transparency
                    doubleSided=True,
#                   extensions={"KHR_materials_unlit": {}},
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 1, 1, 0]
                    )
                )
                #                 pygltflib.Material(
                #                     name="metal",
                #                     doubleSided=True,
                # #                   alphaMode=pygltflib.MASK,
                #                     occlusionTexture=pygltflib.OcclusionTextureInfo(index=1),
                # #                   emissiveFactor=[0.8,0.8,0.8],
                #                     pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                #                         metallicFactor=0.0,
                #                         roughnessFactor=1.0,
                # #                       baseColorFactor=[0.8, 0.8, 0.8, 1],
                #                         baseColorTexture=pygltflib.TextureInfo(index=0),
                #                         metallicRoughnessTexture=pygltflib.TextureInfo(index=1),
                #                     )
                #                 ),
            ],
            bufferViews=[],
            buffers=[pygltflib.Buffer(byteLength=0)],
        )
        self.gltf._glb_data = bytes()


        # Map pairs of (color, alpha) to material's index in material list
        self._color = {(m.name,m.pbrMetallicRoughness.baseColorFactor[3]) if m.pbrMetallicRoughness else (m.name, 0): i
                       for i,m in enumerate(self.gltf.materials)}
        
        self._arrows = {}

        self._node_markers = {}

        #
        # load assets for steel material
        #
        if False:
            for i,file in enumerate(("rust.jpg", "occlusionRoughnessMetallic.png")):
                path  = str(veux.assets/"metal"/file)
                image = pygltflib.Image()
                image.uri = path
                self.gltf.images.append(image)
                self.gltf.textures.append(pygltflib.Texture(source=i, name=path))

            self.gltf.convert_images(pygltflib.ImageFormat.DATAURI)

    def _init_fixed_node(self, points, triangles, x=True, y=True, z=True):
        points_access  = self.position_accessor_index
        indices_access = self.indices_accessor_index

        # Initialize an array for vertex colors, defaulting to a base color
        vertex_colors = np.tile(np.array([0.5, 0.5, 0.5, 1.0], dtype=self.float_t), (len(points), 1))

        # For each triangle, check if the face lies on a plane fixed along given axes
        # and set the corresponding vertices to white if so.
        for tri in triangles:
            # Extract vertices of the triangle
            v0, v1, v2 = points[tri[0]], points[tri[1]], points[tri[2]]

            # Check if all vertices share the same coordinate on any fixed axis
            if x and (np.isclose(v0[0], v1[0]) and np.isclose(v1[0], v2[0])):
                # Set vertices of this triangle to white
                for vi in tri:
                    vertex_colors[vi] = [1.0, 1.0, 1.0, 1.0]
            if y and (np.isclose(v0[1], v1[1]) and np.isclose(v1[1], v2[1])):
                for vi in tri:
                    vertex_colors[vi] = [1.0, 1.0, 1.0, 1.0]
            if z and (np.isclose(v0[2], v1[2]) and np.isclose(v1[2], v2[2])):
                for vi in tri:
                    vertex_colors[vi] = [1.0, 1.0, 1.0, 1.0]


        color_accessor_index = len(self.gltf.accessors)
        self.gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=self._push_data(vertex_colors.astype(self.float_t).tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(vertex_colors),
                type=pygltflib.VEC4,
                max=vertex_colors.max(axis=0).tolist(),
                min=vertex_colors.min(axis=0).tolist(),
            )
        )

        # Create a new material that uses vertex colors directly, e.g., an unlit material.
        # The specifics of creating a material that respects vertex colors depend on your material setup.
        fixed_material = pygltflib.Material(
            name="FixedAxesMaterial",
            pbrMetallicRoughness=pygltflib.PBRMetallicRoughness(
                baseColorFactor=[1.0, 1.0, 1.0, 1.0],  # base color white
                metallicFactor=0.0,
                roughnessFactor=1.0,
            ),
            # Depending on your renderer, you might need to enable vertex colors
        )

        self.gltf.materials.append(fixed_material)
        material_index = len(self.gltf.materials) - 1

        # Create a new mesh primitive using the POSITION and COLOR_0 attributes.
        self.gltf.meshes.append(
            pygltflib.Mesh(
                primitives=[
                    pygltflib.Primitive(
                        mode=pygltflib.TRIANGLES,
                        attributes=pygltflib.Attributes(
                            POSITION=points_access,
                            COLOR_0=color_accessor_index
                        ),
                        material=material_index,
                        indices=indices_access
                    )
                ]
            )
        )
        return len(self.gltf.meshes) - 1


    def _find_marker(self, style: NodeStyle):
        key = (style.color, style.scale)
        if key not in self._node_markers:
            self._node_markers[key] = self._init_marker(style)

        return self._node_markers[key]


    def _init_marker(self, style: NodeStyle):
        #
        #
        #
        index_t = self.index_t
        points = style.scale*np.array(
            [
                [-1.0, -1.0,  1.0],
                [ 1.0, -1.0,  1.0],
                [-1.0,  1.0,  1.0],
                [ 1.0,  1.0,  1.0],
                [ 1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0],
                [ 1.0,  1.0, -1.0],
                [-1.0,  1.0, -1.0],
            ],
            dtype=self.float_t,
        )/10

        triangles = np.array(
            [
                [0, 1, 2],
                [3, 2, 1],
                [1, 0, 4],
                [5, 4, 0],
                [3, 1, 6],
                [4, 6, 1],
                [2, 3, 7],
                [6, 7, 3],
                [0, 2, 5],
                [7, 5, 2],
                [5, 7, 4],
                [6, 4, 7],
            ],
            dtype=index_t,
        )
        triangles_binary_blob = triangles.flatten().tobytes()

        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(triangles_binary_blob,
                                           pygltflib.ELEMENT_ARRAY_BUFFER),
                componentType=GLTF_T[index_t],
                count=triangles.size,
                type=pygltflib.SCALAR,
                max=[int(triangles.max())],
                min=[int(triangles.min())],
            ),
            pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(),
                                           pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        ])


        indices_access = len(self.gltf.accessors)-2
        points_access  = len(self.gltf.accessors)-1
        self.gltf.meshes.append(
               pygltflib.Mesh(
                 primitives=[
                     pygltflib.Primitive(
                         mode=pygltflib.TRIANGLES,
                         attributes=pygltflib.Attributes(POSITION=points_access),
                         material=self._get_material(style),
                         indices=indices_access
                     )
                 ]
               )
        )

        return len(self.gltf.meshes) - 1

    def _init_nodes(self, style: NodeStyle):
        #
        #
        #
        index_t = self.index_t
        points = style.scale*np.array(
            [
                [-1.0, -1.0,  1.0],
                [ 1.0, -1.0,  1.0],
                [-1.0,  1.0,  1.0],
                [ 1.0,  1.0,  1.0],
                [ 1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0],
                [ 1.0,  1.0, -1.0],
                [-1.0,  1.0, -1.0],
            ],
            dtype=self.float_t,
        )/10

        triangles = np.array(
            [
                [0, 1, 2],
                [3, 2, 1],
                [1, 0, 4],
                [5, 4, 0],
                [3, 1, 6],
                [4, 6, 1],
                [2, 3, 7],
                [6, 7, 3],
                [0, 2, 5],
                [7, 5, 2],
                [5, 7, 4],
                [6, 4, 7],
            ],
            dtype=index_t,
        )
        triangles_binary_blob = triangles.flatten().tobytes()

        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(triangles_binary_blob,
                                           pygltflib.ELEMENT_ARRAY_BUFFER),
                componentType=GLTF_T[index_t],
                count=triangles.size,
                type=pygltflib.SCALAR,
                max=[int(triangles.max())],
                min=[int(triangles.min())],
            ),
            pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(),
                                           pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        ])


        indices_access = len(self.gltf.accessors)-2
        points_access  = len(self.gltf.accessors)-1
        self.gltf.meshes.append(
               pygltflib.Mesh(
                 primitives=[
                     pygltflib.Primitive(
                         mode=pygltflib.TRIANGLES,
                         attributes=pygltflib.Attributes(POSITION=points_access),
                         material=self._get_material(style),
                         indices=indices_access
                     )
                 ]
               )
        )

        self._node_mesh = len(self.gltf.meshes) - 1

    def _use_asset(self, name, scale, rotation, material):
        pass

    def _get_material(self, style: DrawStyle)->int:
        if hasattr(style,"alpha"):
            alpha = style.alpha
        else:
            alpha = 1.0

        color = style.color

        rgb = None
        if (color, alpha) in self._color:
            return self._color[(color,alpha)]

        elif isinstance(color, str) and color[0] == "#":
            # Remove leading hash in hex
            hx  = color.lstrip("#")
            # Convert hex to RGB
            rgb = [int(hx[i:i+2], 16)/255 for i in (0, 2, 4)]

        elif isinstance(color, tuple):
            rgb = color

        elif isinstance(color, str):
            for key,a in self._color:
                if color == key:
                    rgb = self.gltf.materials[self._color[(key,a)]].pbrMetallicRoughness.baseColorFactor[:3]

        if rgb is None:
            raise TypeError("Unexpected type for color")

        # Store index for new material
        self._color[(color, alpha)] = len(self.gltf.materials)
        # Create and add new material
        self.gltf.materials.append(
            pygltflib.Material(
                name=str(color),
                doubleSided=True,
                alphaMode=pygltflib.BLEND,
                pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                    baseColorFactor=[*rgb, alpha]
                )
            ),
        )
        return self._color[(color, alpha)]


    def _push_data(self, data, target=None, byteStride=None)->int:
        self.gltf.bufferViews.append(
                pygltflib.BufferView(
                    buffer=0,
                    byteStride=byteStride,
                    byteOffset=self.gltf.buffers[0].byteLength,
                    byteLength=len(data),
                    target=target,
                )
        )

        self.gltf._glb_data += data
        self.gltf.buffers[0].byteLength += len(data)
        return len(self.gltf.bufferViews)-1


    def plot_nodes(self, vertices, label = None, style=None, data=None, names=None, rotations=None, skin=False, **kwds):
        if names is None:
            names = itertools.repeat(None)

        nodes = []

        # if not hasattr(self, "_node_mesh"):
        #     self._init_nodes(style or NodeStyle())

        marker = self._find_marker(style or NodeStyle())

        if rotations is None:
            rotations = itertools.repeat([0, 0, 0, 1.0])
        else:
            try:
                rotations = Rotation.from_matrix([self._rotation_matrix@R for R in rotations]).as_quat().tolist()
            except ValueError:
                rotations = Rotation.from_rotvec([self._rotation_matrix@R for R in rotations]).as_quat().tolist()


        for name,coord, rotation in zip(names, vertices, rotations):
            index = _append_index(self.gltf.nodes, pygltflib.Node(
                    mesh=marker,
                    name=str(name),
                    rotation=rotation,
                    translation=(self._rotation_matrix@coord).tolist(),
                )
            )
            if not skin:
                self.gltf.scenes[0].nodes.append(index)
            nodes.append(Node(id=index))

        return nodes


    def add_lines(self, lines: list, style=None, skin_nodes=None):
        """
        Add skinned lines to the glTF object that connect pairs of nodes specified in `lines`.
        The lines will deform as the corresponding nodes are translated.

        :param lines: A list of pairs of indices (i, j), where i and j are node indices.
        :param access_vertices: The accessor index for initial positions of the nodes.
        """
        gltf = self.gltf
        scene = gltf.scenes[0]
        EYE4 = np.eye(4, dtype=self.float_t)

        if len(lines) == 0:
            return

        # Validate that all node indices in `lines` exist in gltf.nodes
        max_node_idx = len(gltf.nodes) - 1
        for i, j in lines:
            if i > max_node_idx or j > max_node_idx:
                raise ValueError(f"Node indices {i} or {j} in `lines` are out of range for `gltf.nodes`.")


        # Create joints (one per node) and bind the lines to them
        if skin_nodes is not None:
            joint_nodes = list({skin_nodes[idx].id for pair in lines for idx in pair})  # Unique joint indices
        else:
            joint_nodes = list({idx for pair in lines for idx in pair})  # Unique joint indices

        joint_node_to_index = {node: i for i, node in enumerate(joint_nodes)}

        points = np.zeros((len(joint_nodes), 3), dtype=self.float_t)

        ver_accessor = _append_index(self.gltf.accessors, pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        )
    
        # Create skeleton
        skeleton_root_node = pygltflib.Node(name="LineSkeleton", 
                                            children=joint_nodes)
        skeleton_root_idx = _append_index(gltf.nodes, skeleton_root_node)
        scene.nodes.append(skeleton_root_idx)

        # Inverse bind matrices
        inverse_bind_matrices = []
        for joint_node in joint_nodes:
            # Compute the transform matrix for the bind pose
            node = gltf.nodes[joint_node]
            t_matrix = np.eye(4, dtype=self.float_t)
            t_matrix[:3, 3] = node.translation if node.translation else [0.0, 0.0, 0.0]
            rotation_matrix = Rotation.from_quat(node.rotation).as_matrix() if node.rotation else EYE3
            t_matrix[:3, :3] = rotation_matrix #* (node.scale if node.scale else 1.0)
            # t_matrix = np.linalg.inv(t_matrix)

            # inverse_bind_matrices.append(np.array(t_matrix, dtype=self.float_t).T)
            inverse_bind_matrices.append(EYE4)

        # Add buffer view and accessor for the inverse bind matrices
        # Flatten inverse bind matrices for glTF format
        ibm_array = np.array(inverse_bind_matrices, dtype=self.float_t).reshape(-1, 16)

        ibm_accessor_idx = _append_index(gltf.accessors, pygltflib.Accessor(
            bufferView=self._push_data(ibm_array.tobytes(), target=None),
            componentType=GLTF_T[self.float_t],
            count=len(joint_nodes),
            type="MAT4"
        ))

        # Create the skin
        if not gltf.skins:
            gltf.skins = []
        skin_idx = len(gltf.skins)
        gltf.skins.append(pygltflib.Skin(
            inverseBindMatrices=ibm_accessor_idx,
            joints=joint_nodes,
            skeleton=skeleton_root_idx,  # Root joint
            name="LineSkin"
        ))

        # Create the index buffer for the lines
        # indices = []
        joints_0 = []
        weights_0 = []
        for i, j in lines:
            # indices.extend([i, j])
            joints_0.extend([
                [joint_node_to_index[i], 0, 0, 0],
                [joint_node_to_index[j], 0, 0, 0]
            ])
            weights_0.extend([
                [1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0]
            ])

        index_array = np.array(lines, dtype=self.index_t).reshape(-1)

        # Add buffer view and accessor

        index_accessor_idx = len(gltf.accessors)
        gltf.accessors.append(pygltflib.Accessor(
            bufferView=self._push_data(index_array.tobytes(),pygltflib.ELEMENT_ARRAY_BUFFER),
            componentType=GLTF_T[self.index_t],
            count=len(index_array),
            type="SCALAR",
            min=[int(index_array.min())],
            max=[int(index_array.max())]
        ))

        # Create the line mesh

        joints_0_accessor_idx = len(gltf.accessors)
        gltf.accessors.append(pygltflib.Accessor(
            bufferView=self._push_data(np.array(joints_0, dtype=self.index_t).tobytes(), pygltflib.ARRAY_BUFFER),
            componentType=GLTF_T[self.index_t],
            count=len(joints_0),
            type="VEC4"
        ))

        weights_0_array = np.array(weights_0, dtype=self.float_t)
        weights_0_accessor_idx = len(gltf.accessors)
        gltf.accessors.append(pygltflib.Accessor(
            bufferView=self._push_data(weights_0_array.tobytes(), pygltflib.ARRAY_BUFFER),
            componentType=GLTF_T[self.float_t],
            count=len(weights_0),
            type="VEC4"
        ))

        # Create the line mesh
        mesh_idx = len(gltf.meshes)
        line_mesh = pygltflib.Mesh(
            primitives=[
                pygltflib.Primitive(
                    attributes=pygltflib.Attributes(
                        POSITION=ver_accessor,
                        JOINTS_0=joints_0_accessor_idx,
                        WEIGHTS_0=weights_0_accessor_idx
                    ),
                    indices=index_accessor_idx,
                    mode=pygltflib.LINES,
                    material=self._get_material(style or LineStyle())
                )
            ], name="LineSkinMesh")
        gltf.meshes.append(line_mesh)

        # Create a node for the line mesh
        mesh_node = _append_index(gltf.nodes, pygltflib.Node(
            mesh=mesh_idx, 
            skin=skin_idx, 
            # rotation=self._rotation,
            name="LineSkinMeshNode"
        ))

        # Add the new node to the scene
        scene.nodes.append(mesh_node)

    def set_data(self, data, key):

        points = np.array(data, dtype=self.float_t)
        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        ])

        self._data[key] = {"access_index": len(self.gltf.accessors)-1}


    def _make_line_strip(self, indices, vertices, points_access, material):
        
        # `n` adjusts indices by the number of nan rows that were removed so far
        if not isinstance(vertices, (int, str)):
            n  = sum(np.isnan(vertices[:indices[0],0]))
        else:
            n = 0

        indices_array = indices - np.dtype(self.index_t).type(n)

        indices_binary_blob = indices_array.tobytes()

        if len(indices_array) <= 1:
            import warnings
            warnings.warn(indices_array)
            return None

        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(indices_binary_blob,
                                            pygltflib.ELEMENT_ARRAY_BUFFER),
                componentType=GLTF_T[self.index_t],
                count=indices_array.size,
                type=pygltflib.SCALAR,
                max=[int(indices_array.max())],
                min=[int(indices_array.min())],
            )
        ])
        self.gltf.meshes.append(
                pygltflib.Mesh(
                    primitives=[
                        pygltflib.Primitive(
                            mode=pygltflib.LINE_STRIP,
                            attributes=pygltflib.Attributes(POSITION=points_access),
                            material=material,
                            # most recently added accessor
                            indices=len(self.gltf.accessors)-1,
                            # TODO: use this mechanism to add annotation data
                            extras={},
                        )
                    ]
                )
        )

        self.gltf.nodes.append(pygltflib.Node(
                mesh=len(self.gltf.meshes)-1,
                rotation=self._rotation
            )
        )

        self.gltf.scenes[0].nodes.append(
            len(self.gltf.nodes)-1
        )

        return Line(id=len(self.gltf.nodes)-1)

    def plot_lines(self, vertices, indices=None, style: LineStyle | None = None, **node_kwds):
        """
        Add a batch of disconnected line-segments to the scene as a *single*
        glTF primitive (mode = LINES).

        Parameters
        ----------
        vertices : array_like, int, or str
            • (N, 3) array – NaNs may delimit poly-lines  
            • accessor index/name – reuse an existing POSITION accessor
        indices : sequence[array_like] | None
            If given, each item is a vertex-index list describing one poly-line.
            Consecutive pairs are turned into independent segments.  If omitted
            and ``vertices`` is an array with NaNs, the index list is generated
            automatically.
        style   : LineStyle, optional
            Passed to ``_get_material``.
        **node_kwds
            Extra kwargs for the created glTF ``Node`` (e.g. “name”, “matrix”, ...).

        Returns
        -------
        int
            Index of the new ``Node`` in ``self.gltf.nodes``.
        """
        #
        # 1.  Material
        #
        material_idx = self._get_material(style or LineStyle())

        # ------------------------------------------------------------------
        # 2.  POSITION accessor
        # ------------------------------------------------------------------
        if isinstance(vertices, (int, str)):
            # Re-use an accessor that caller has already stored.
            points_access = self.get_data(vertices)["access_index"]
            n_vertices = self.gltf.accessors[points_access].count
            # When the caller passes indices, we trust them to
            # reference this existing accessor correctly.
            verts_map = None          # identity
        else:
            return self.plot_lines_old(vertices, indices, style, **node_kwds)
            vertices = np.asarray(vertices, dtype=self.float_t)
            if vertices.ndim != 2 or vertices.shape[1] != 3:
                raise ValueError("vertices must be (N,3)")

            # Strip NaN rows – these are *delimiters*, never emitted.
            finite_mask = np.isfinite(vertices[:, 0])
            points = vertices[finite_mask]
            if points.size == 0:
                return None  # nothing to draw

            buf_view = self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER)
            self.gltf.accessors.append(
                pygltflib.Accessor(
                    bufferView=buf_view,
                    componentType=GLTF_T[self.float_t],
                    count=len(points),
                    type=pygltflib.VEC3,
                    min=points.min(axis=0).tolist(),
                    max=points.max(axis=0).tolist(),
                )
            )
            points_access = len(self.gltf.accessors) - 1
            n_vertices = len(points)

            # Map original row index to new compact index
            verts_map = np.full(len(vertices), -1, dtype=self.index_t)
            verts_map[finite_mask] = np.arange(n_vertices, dtype=self.index_t)

        # ------------------------------------------------------------------
        # 3.  Build flat SCALAR index list (pairs)
        # ------------------------------------------------------------------
        seg_idx = []
        if indices is not None:
            for poly in indices:
                poly = np.asarray(poly, dtype=self.index_t)
                if verts_map is not None:
                    poly = verts_map[poly]          # translate to compact space
                if len(poly) < 2:
                    continue
                # consecutive pairs
                for i in range(len(poly) - 1):
                    a, b = poly[i], poly[i + 1]
                    if a != b:                      # skip degenerate
                        seg_idx.extend((a, b))
        else:
            if verts_map is None:
                raise ValueError("Automatic index generation needs vertices array.")
            current = []
            for src_idx, compact_idx in enumerate(verts_map):
                if compact_idx == -1:               # NaN delimiter
                    current.clear()
                    continue
                current.append(compact_idx)
                if len(current) >= 2:
                    # last two points form a segment
                    seg_idx.extend(current[-2:])
        if not seg_idx:
            return None

        seg_idx = np.asarray(seg_idx, dtype=self.index_t)
        if seg_idx.max() >= n_vertices:
            raise ValueError("Index out of range for POSITION accessor.")


        # ------------------------------------------------------------------
        # Create lines
        # ------------------------------------------------------------------

        #
        # INDICES accessor
        # 
        idx_view = self._push_data(seg_idx.tobytes(), pygltflib.ELEMENT_ARRAY_BUFFER)
        self.gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=idx_view,
                componentType=GLTF_T[self.index_t],
                count=seg_idx.size,
                type=pygltflib.SCALAR,
                min=[int(seg_idx.min())],
                max=[int(seg_idx.max())],
            )
        )
        idx_acc = len(self.gltf.accessors) - 1

        #
        # Primitive, Mesh, and Node
        #
        primitive = pygltflib.Primitive(
            attributes={"POSITION": points_access},
            indices=idx_acc,
            mode=pygltflib.LINES,
            material=material_idx,
        )
        self.gltf.meshes.append(pygltflib.Mesh(primitives=[primitive]))
        mesh_idx = len(self.gltf.meshes) - 1

        self.gltf.nodes.append(pygltflib.Node(mesh=mesh_idx, **node_kwds))
        node_idx = len(self.gltf.nodes) - 1

        self.gltf.scenes[0].nodes.append(node_idx)

        return Line(id=node_idx)

    def plot_lines_old(self, vertices, indices=None, style: LineStyle=None, **kwds):

        lines = []
        material = self._get_material(style or LineStyle())

        points_access = None 

        if isinstance(vertices, (int, str)):
            points_access = self.get_data(vertices)["access_index"]

        else:
            # vertices is given with nans separating line groups, but
            # GLTF does not accept nans so we have to filter these
            # out, and add distinct meshes for each line group
            assert np.all(np.isnan(vertices[np.isnan(vertices[:,0]), :]))
            points  = np.array(vertices[~np.isnan(vertices[:,0]),:], dtype=self.float_t)

            if points.size == 0:
                return

            self.gltf.accessors.append(
                pygltflib.Accessor(
                    bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                    componentType=GLTF_T[self.float_t],
                    count=len(points),
                    type=pygltflib.VEC3,
                    max=points.max(axis=0).tolist(),
                    min=points.min(axis=0).tolist(),
                )
            )
            points_access = len(self.gltf.accessors) - 1


        if indices is None:
            # Get indices by splitting at nans
            indices_ = utility.split(np.arange(len(vertices), dtype=self.index_t), np.nan, vertices[:,0])
        else:
            indices_ = list(map(lambda x: np.array(x, dtype=self.index_t), indices))

        for indices in indices_:
            if (line := self._make_line_strip(indices, vertices, points_access, material)) is None:
                continue
            lines.append(line)

        return lines

    def plot_vectors(self, locs, vecs, label=None, extrude=False, **kwds):
        ne = len(vecs)
        if not extrude:
            for j in range(3):
                style = kwds.get("line_style", LineStyle(color=("red", "green", "blue")[j]))
                X = np.zeros((ne*3, 3))*np.nan
                for i in range(j,ne,3):
                    X[i*3,:] = locs[i]
                    X[i*3+1,:] = locs[i] + vecs[i]
                self.plot_lines(X, style=style, label=label)
        else:
            for j in range(3):
                style = kwds.get("line_style", LineStyle(color=("red", "green", "blue")[j]))
                X = np.zeros((ne*3, 3))*np.nan
                for i in range(j,ne,3):
                    self.draw_arrow(locs[i], vector=vecs[i], style=style)


    def draw_arrow(self, location, rotation=None, size=None, vector=None, style=None):
        from veux.assets import create_arrow

        def quaternion_from_x_to_vec(v):
            """
            Returns a rotation that rotates [1,0,0] onto the vector v.
            """
            # Convert to float np.array and normalize
            v = np.array(v, dtype=float)
            if np.allclose(v, 0):
                raise ValueError("Target vector must be non-zero.")
            v_norm = v / np.linalg.norm(v)

            # The reference "from" vector is the unit x-axis
            x_axis = np.array([1.0, 0.0, 0.0])

            # Check if they are already the same (no rotation) or exactly opposite
            dot_val = np.dot(x_axis, v_norm)
            if np.isclose(dot_val, 1.0):
                # Same direction, set to identity quaternion
                return Rotation.from_quat([0.0, 0.0, 0.0, 1.0])
            elif np.isclose(dot_val, -1.0):
                # Opposite direction, 180 deg rotation about any axis perpendicular to x
                # e.g., about [0,1,0]
                return Rotation.from_rotvec(np.pi * np.array([0, 1, 0]))

            # Otherwise, compute rotation axis as cross(x_axis, v_norm) and rotation angle
            axis = np.cross(x_axis, v_norm)
            axis /= np.linalg.norm(axis)
            angle = np.arccos(dot_val)

            return Rotation.from_rotvec(angle * axis)

        if style is None:
            style = DrawStyle(color="red")

        if vector is not None:
            size = np.linalg.norm(vector)
            rotation = quaternion_from_x_to_vec(vector/size).as_quat().tolist()

        if style.color not in self._arrows:
            self._arrows[style.color] = create_arrow(self, size, style)

        arrow = self._arrows[style.color]

        index = _append_index(self.gltf.nodes, pygltflib.Node(
                mesh=arrow,
                rotation=rotation,
                translation=(self._rotation_matrix@location).tolist(),
            )
        )
        self.gltf.scenes[0].nodes.append(index)

    
    def draw_skin(self, vertices, triangles):
        pass

    def plot_mesh(self, vertices, triangles, local_coords=None, style=None,
                  joints_0=None, 
                  weights_0=None,
                  skin=None,
                  node_name=None,
                  mesh_name=None,
                  **kwds) -> tuple:

        material  = self._get_material(style or MeshStyle())

        if isinstance(triangles, int):
            index_access = triangles
        else:
            triangles = np.array(triangles, dtype=self.index_t)
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(triangles.flatten().tobytes(), pygltflib.ELEMENT_ARRAY_BUFFER),
                    componentType=GLTF_T[self.index_t],
                    count=triangles.size,
                    type=pygltflib.SCALAR,
                    max=[int(triangles.max())],
                    min=[int(triangles.min())],
                )
            ])
            index_access = len(self.gltf.accessors)-1

        if isinstance(vertices, int):
            point_access = vertices
        else:
            points    = np.array(vertices, dtype=self.float_t)
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                    componentType=GLTF_T[self.float_t],
                    count=len(points),
                    type=pygltflib.VEC3,
                    max=points.max(axis=0).tolist(),
                    min=points.min(axis=0).tolist(),
                )
            ])
            point_access = len(self.gltf.accessors)-1

#       # Expecting morph_targets as an iterable of target arrays (each of shape (n,3))
#       morph_targets = kwds.get("morph_targets", None)
#       targets = []
#       if morph_targets is not None:
#           for mt in morph_targets:
#               mt_arr = np.array(mt, dtype=self.float_t)
#               self.gltf.accessors.extend([
#                   pygltflib.Accessor(
#                       bufferView=self._push_data(mt_arr.tobytes(), pygltflib.ARRAY_BUFFER),
#                       componentType=GLTF_T[self.float_t],
#                       count=len(mt_arr),
#                       type=pygltflib.VEC3,
#                       max=mt_arr.max(axis=0).tolist(),
#                       min=mt_arr.min(axis=0).tolist(),
#                   )
#               ])
#               mt_accessor = len(self.gltf.accessors)-1
#               targets.append({"POSITION": mt_accessor})

        mesh = pygltflib.Mesh(
                primitives=[
                    pygltflib.Primitive(
                        mode=pygltflib.TRIANGLES,
                        attributes=pygltflib.Attributes(
                            POSITION=point_access
                        ),
                        material=material,
                        indices=index_access,
                     )
                ]
            )

        if mesh_name is not None:
            mesh.name = mesh_name

        if joints_0 is not None:
            joints_0  = np.array(joints_0,  dtype="uint16")
            mesh.primitives[0].attributes.JOINTS_0 = _append_index(self.gltf.accessors, pygltflib.Accessor(
                bufferView=self._push_data(joints_0.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T["uint16"],
                count=len(joints_0),
                type="VEC4"
            ))

        if weights_0 is not None:
            weights_0 = np.array(weights_0, dtype=self.float_t)
            mesh.primitives[0].attributes.WEIGHTS_0 = _append_index(self.gltf.accessors, pygltflib.Accessor(
                bufferView=self._push_data(weights_0.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(weights_0),
                type="VEC4"
            ))


        if local_coords is not None:
            locoor = np.array(local_coords, dtype=self.float_t)
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(locoor.tobytes(), pygltflib.ARRAY_BUFFER),
                    componentType=GLTF_T[self.float_t],
                    count=len(locoor),
                    type=pygltflib.VEC2,
                    max=locoor.max(axis=0).tolist(),
                    min=locoor.min(axis=0).tolist(),
                )
            ])
            mesh.primitives[0].attributes.TEXCOORD_0 = len(self.gltf.accessors) -1
        
        #
        # 4) Create a Node referencing the mesh + skin
        #
        self.gltf.nodes.append(pygltflib.Node(
                mesh=_append_index(self.gltf.meshes, mesh),
                rotation=self._rotation
            )
        )
        if node_name is not None:
            self.gltf.nodes[-1].name = node_name

        if skin is not None:
            self.gltf.nodes[-1].skin = skin

        # Add to scene
        scene_node = len(self.gltf.nodes)-1
        self.gltf.scenes[0].nodes.append(scene_node)

        return Mesh(id=scene_node,
                    vertices=point_access, 
                    indices=index_access)


    def plot_mesh_field(self, mesh_handle, field,
                        colormap="rainbow4", # "cet_CET_R1",# "cet_CET_D13"#"twilight", #"viridis", 
                        vmin=None, vmax=None,
                        **kwds) -> tuple:
        """
        Draw a mesh colored by a scalar field.
        
        :param vertices: Nx3 array of vertex positions
        :param triangles: Mx3 array of triangle indices
                          (or an int referencing an existing accessor)
        :param field: length-N array (one scalar per vertex)
        :param style: optional style info
        :param colormap: name of a matplotlib colormap or something similar
        :param vmin, vmax: optionally set the data range for mapping
        :param kwds: other keywords
        """

        # Determine color array from field
        color_array = self._map_field_to_colors(field, colormap, vmin, vmax)

        color_array = color_array.astype(self.float_t)
        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(color_array.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(color_array),
                type=pygltflib.VEC4,  # RGBA
                max=color_array.max(axis=0).tolist(),
                min=color_array.min(axis=0).tolist(),
            )
        ])
        color_access = len(self.gltf.accessors) - 1

        # Attach the color accessor to the existing mesh’s attributes
        mesh_idx = self.gltf.nodes[mesh_handle.id].mesh
        self.gltf.meshes[mesh_idx].primitives[0].attributes.COLOR_0 = color_access

        return mesh_handle

    def _map_field_to_colors(self, field, 
                             colormap="cet_CET_D13", #"rainbow4", #"viridis", 
                             vmin=None, vmax=None):
        """
        """
        try:
            import colorcet as cc
            field = np.asarray(field, dtype=np.float64)

            # Normalisation
            if vmin is None:
                vmin = np.nanmin(field)
            if vmax is None:
                vmax = np.nanmax(field)

            # avoid divide-by-zero
            span = vmax - vmin or 1.0
            t = np.clip((field - vmin) / span, 0.0, 1.0)


            try:
                cmap = cc.cm[colormap]
            except KeyError as e:
                raise ValueError(f"Unknown ColorCET colormap '{colormap}'.") from e

            colors = cmap(t)
            return colors

        except:
            import matplotlib
            import matplotlib.cm

            field = np.array(field, dtype=np.float64)
            if vmin is None:
                vmin = field.min()
            if vmax is None:
                vmax = field.max()

            norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
            cmap = matplotlib.cm.get_cmap(colormap)

            colors = cmap(norm(field))
            return colors


    def to_glb(self)->bytes:
        return b"".join(self.gltf.save_to_bytes())

    def write(self, filename=None):

        self.gltf.save(filename)

#       if "glb" in filename[-3:]:
#           glb = b"".join(self.gltf.save_to_bytes())
#           with open(filename,"wb+") as f:
#               f.write(glb)




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio M. Perez
#
# Summer 2024
#
import os
import base64
import textwrap
from pathlib import Path
import numpy as np
import numpy as np
from io import BytesIO

class Viewer:
    """
    A class to represent a 3D model viewer.

    Methods:
    --------
    __init__(self, viewer=None, path=None, data=None):
        Initializes the Viewer with optional viewer type, file path, or binary data.

    """
    def __init__(self, thing,
                 viewer=None, id=None, 
                 plane = False,
                 size=None,
                 hosted=None,
                 show_quit=True,
                 quit_on_load=False,
                 standalone=True,
                 lights=None):
        
        """
        If hosted is True, the viewier will fetch the model.

        If standalone is True, the viewer will be rendered as a standalone HTML page.
        """
        self._id = id if id is not None else "veux-viewer"
        self._viewer = viewer if viewer is not None else os.environ.get("VEUX_VIEWER","mv")
        self._lights = lights # light or dark mode
        self._plane = plane
        self._size = size
        self._show_quit = show_quit
        self._quit_on_load = quit_on_load

        self._hosted = hosted
        if hosted is None and self._viewer == "mv":
            self._hosted = False
        elif hosted is None:
            self._hosted = True

        if hasattr(thing, "canvas"):
            # artist was passed
            canvas = thing.canvas
            data = canvas.to_glb()
        elif hasattr(thing, "to_glb"):
            canvas = thing
            data = canvas.to_glb()
        else:
            data = thing

        self._standalone = standalone

        if not self._hosted:
            self._model_data = None
            data64 = base64.b64encode(data).decode('utf-8')
            self._glbsrc=f"data:model/gltf-binary;base64,{data64}"
        else:
            self._model_data = data
            self._glbsrc = "/model.glb" 


    def resources(self):
        if self._hosted:
            yield ("/model.glb", lambda : self._model_data)

            if self._viewer == "mv":
                yield ("/black_ground.hdr", _serve_black_ground_hdr)


    def get_html(self):
        if self._viewer == "babylon":
            with open(Path(__file__).parents[0]/"babylon.html", "r") as f:
                return f.read()

        if self._viewer in {"three-170", "three"}:
            with open(Path(__file__).parents[0]/"three-170.html", "r") as f:
                return f.read()

        if self._viewer == "three-160":
            with open(Path(__file__).parents[0]/"three-160.html", "r") as f:
                return f.read()

        elif self._viewer == "three-130":
            with open(Path(__file__).parents[0]/"three-130.html", "r") as f:
                return f.read()

        elif self._viewer == "mv":
            return _model_viewer(self._glbsrc,
                                 control=True,
                                 plane=self._plane,
                                 size=self._size,
                                 show_quit=self._show_quit,
                                 quit_on_load=self._quit_on_load,
                                 hosted=self._hosted,
                                 light_mode=self._lights,
                                 standalone=self._standalone)


class VeuxTable:
    def html(self):
        with open(Path(__file__).parents[0]/"mv/veux-select.html", "r") as f:
            return f.read()
    def css(self):
        with open(Path(__file__).parents[0]/"mv/veux-select.css", "r") as f:
            return f.read()
    def js(self):
        with open(Path(__file__).parents[0]/"mv/veux-select.js", "r") as f:
            return f.read()

class EmptyObject:
    def html(self):
        return ""
    def css(self):
        return ""
    def js(self):
        return ""

def _model_viewer(source,
                  control=False,
                  size=None,
                  hosted=False,
                  plane=False,
                  show_quit=None,
                  quit_on_load=False,
                  standalone=True,
                  light_mode=None):

    if light_mode is None:
        light_mode = "light"

    lights = """
      <style>
      html {
        color-scheme: light; /* dark */
        background-color: #555555;
        --poster-color: #555555;
        --progress-bar-color: #fff;
        --progress-bar-background-color: rgba(255, 255, 255, 0.2);
      }
      @media (prefers-color-scheme: dark) {
        model-viewer {
          --poster-color: #555555;
          background-color: #555555;
        }
      }
      </style>
    """
    lights = ""

    with open(Path(__file__).parents[0]/"controls.css", "r") as f:
        control_style = f"<style>{f.read()}</style>"

    with open(Path(__file__).parents[0]/"controls.js", "r") as f:
        control_script = f"<script>{f.read()}</script>"

    with open(Path(__file__).parents[0]/"model-viewer.min.js", "r") as f:
        library = f'<script type="module">{f.read()}</script>'


    control_html = """
      <div class="controls">
        <button id="toggle-animation">Pause</button>
      </div>
    """

    
    if hosted:
        environment = "/black_ground.hdr"
    else:
        environment = "neutral"
    
    if False: #plane:
        camera = """
            camera-controls
            disable-rotate
            camera-orbit="0deg 90deg auto"
            auto-rotate="false"
        """
    else:
        camera = """
            camera-controls
            max-field-of-view="90deg" 
        """
            # min-camera-orbit='auto auto 100%'
            # max-camera-orbit='auto auto 100%'
            # min-camera-orbit="auto auto 0m"

    if size is None:
        size = 'style="width: 100%; height: 100vh;"'
    else:
        size = f'style="width: {size[0]}px; height: {size[1]}px;"'
    
    quit_button = ""
    if show_quit:
        quit_button = """<form action="/quit" method="get" target="quit-frame" style="display:inline;">
            <button type="submit" style="
                display:inline-flex;align-items:center;justify-content:center;
                width:24px;height:24px;border:none;border-radius:4px;
                background:#eee;color:#333;text-decoration:none;
                font-weight:bold;font-size:16px;
                transition:background 0.2s ease;cursor:pointer;"
                onmouseover="this.style.background='#ccc'"
                onmouseout="this.style.background='#eee'">
                <svg viewBox="0 0 10 10" width="12" height="12" fill="none" stroke="currentColor" stroke-width="2">
                <line x1="1" y1="1" x2="9" y2="9" />
                <line x1="9" y1="1" x2="1" y2="9" />
                </svg>
            </button>
          </form>
          <iframe name="quit-frame" style="display:none;"></iframe>
        """
    
    quit_script = ""
    if quit_on_load:
        quit_script = """
        <script>
        window.addEventListener('load', function() {
            const quitButton = document.querySelector('form[action="/quit"] button');
            if (quitButton) {
                quitButton.click();
            }
        });
        </script>
        """

    capt_button = """<button id="capture-button">Capture</button>"""
    capt_script =  """
    <script>
    const modelViewer = document.getElementById("veux-viewer");

    async function downloadPosterToBlob() {
        const blob = await modelViewer.toBlob({ idealAspect: false });
        const url = URL.createObjectURL(blob);
        const a = document.createElement("a");
        a.href = url;
        a.download = "modelViewer_toBlob.png";
        a.click();
        URL.revokeObjectURL(url);
    }

    function downloadPosterToDataURL() {
        const url = modelViewer.toDataURL();
        const a = document.createElement("a");
        a.href = url;
        a.download = "modelViewer_toDataURL.png";
        a.click();
        URL.revokeObjectURL(url);
    }
    document.querySelector("#capture-button").addEventListener("click", downloadPosterToBlob);
    
    </script>
    """

    table = EmptyObject() # VeuxTable()

    viewer = f"""
          {quit_button}
          {capt_button if show_quit else ""}
          <model-viewer 
            id="veux-viewer"
            alt="rendering"
            src="{source}"
            autoplay
            {size}
            max-pixel-ratio="2"
            interaction-prompt="none"
            shadow-intensity="1"
            environment-image="{environment}"
            shadow-light="10000 10000 10000"
            exposure="0.8"
            {camera}
            touch-action="pan-y">
            <div class="progress-bar hide" slot="progress-bar">
                <div class="update-bar"></div>
            </div>
          </model-viewer>
    """

    if not standalone:
        page = f"""
        <div style='display: flex; flex-direction: row;'>
        {viewer}
        </div>
        {library}
        """
    else:
        page = f"""
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
            "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
            <head>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <title>veux</title>
            {library}
            {lights}
            {control_style}
            <style>{table.css()}</style>
            </head>
            <body>
                {viewer}
                {table.html()}
                {control_html if control else ""}
                {control_script if control else ""}
            </body>
            {quit_script}
            {capt_script}
            <script>{table.js()}</script>
        </html>
        """
    return textwrap.dedent(page)


def _serve_black_ground_hdr():
    import bottle
    width, height = 1024, 512

    # Create a blank HDR image
    hdr_image = np.ones((height, width, 3), dtype=np.float32)  # Start with white
    horizon = int(height * 0.6)
    hdr_image[horizon:, :] = 0.0  # Black ground

    # Create the HDR header
    hdr_header = (
        "#?RADIANCE\n"
        "FORMAT=32-bit_rle_rgbe\n\n"
        f"-Y {height} +X {width}\n"
    )

    # Convert the RGB values to Radiance RGBE format
    rgbe_image = np.zeros((height, width, 4), dtype=np.uint8)
    brightest = np.maximum.reduce(hdr_image, axis=2)
    nonzero_mask = brightest > 0
    mantissa, exponent = np.frexp(brightest[nonzero_mask])
    rgbe_image[nonzero_mask, :3] = (hdr_image[nonzero_mask] / mantissa[:, None] * 255).astype(np.uint8)
    rgbe_image[nonzero_mask, 3] = (exponent + 128).astype(np.uint8)

    # Encode the HDR data to memory
    hdr_data = BytesIO()
    hdr_data.write(hdr_header.encode('ascii'))  # Write the header
    hdr_data.write(rgbe_image.tobytes())  # Write the pixel data

    # Serve the HDR file
    return bottle.HTTPResponse(
        body=hdr_data.getvalue(),
        status=200,
        headers={"Content-Type": "image/vnd.radiance"}
    )


#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Copyright (c) 2025, Claudio M. Perez
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#
from collections import defaultdict

import numpy as np
from scipy.spatial.transform import Rotation

import pygltflib
from pygltflib import FLOAT

from veux.canvas.gltf import GLTF_T
from veux.config import MeshStyle, LineStyle

from veux.frame.extrude import ExtrusionCollection, add_extrusion
from shps.frame.extrude import FrameMesh
from veux.frame._element import _SimpleSample

def _append_index(lst, item):
    lst.append(item)
    return len(lst) - 1


def skin_frames(model, artist, config=None, interpolate=None):
    """
    Build a skinned mesh for all frame elements in the reference (undeformed) configuration.
    Returns a dictionary mapping (element_name, j) -> glTF node index
    """
    if config is None:
        config = {
            "style": MeshStyle(color="gray"),
            "scale": 1.0,
            "outline": "",
        }
    scale = config.get("scale", 1.0)
    canvas = artist.canvas 
    Ra = artist._plot_rotation


    #
    # Create a skeleton root node
    #
    gltf = canvas.gltf
    skeleton_root_node = pygltflib.Node(name="FrameExtrusionSkeletonRoot", children=[])
    skeleton_root_idx = _append_index(gltf.nodes, skeleton_root_node)
    gltf.scenes[0].nodes.append(skeleton_root_idx)

    #
    joint_nodes = skeleton_root_node.children
    ibms = []
    skin_nodes = {}
    joint_elements = []

    def _bind_inv(translation, rotmat):
        M = np.eye(4, dtype=canvas.float_t)
        M[:3,:3] = rotmat
        M[:3, 3] = translation
        return np.linalg.inv(M).T

    #
    # 3) For each ring, create a glTF Node (joint),
    #    and assign ring vertices to that joint
    #
    I = 0
    joints_0    = []
    weights_0   = []
    e = ExtrusionCollection([], [], [], set(), set())
    for tag in model.iter_cell_tags():
        if not model.cell_matches(tag, "frame") and not model.cell_matches(tag, "truss"):
            continue
        
        if False:
            X = np.array([Ra@model.node_position(n) for n in model.cell_nodes(tag)])
            ns = len(X)
            R = [Ra@model.frame_orientation(tag).T]*ns

            sections = [model.frame_section(tag, i) for i in range(ns)]
        else:
            elem = model.frame_element(tag)
            X = np.array([
                Ra@elem.sample_position(i) for i in elem.simple_samples()
            ])
            ns = len(X)
            R = [
                Ra@elem.sample_rotation(i) for i in elem.simple_samples()
            ]
            sections = [elem.sample_section(i) for i in elem.simple_samples()]

        if sections[0] is None or sections[-1] is None:
            continue

        extr = FrameMesh(len(X),
                        [s.exterior() for s in sections],
                        scale=scale,
                        do_end_caps=False)

        I += add_extrusion(extr, e, X, R, I)

        for j, start_idx, end_idx in extr.ring_ranges():
            #
            node = pygltflib.Node()
            node.translation =  X[j].tolist()
            node.rotation    =  Rotation.from_matrix(R[j]).as_quat().tolist()

            skin_nodes[(tag, j)] = _append_index(gltf.nodes, node)

            # add to skeleton root
            joint = _append_index(joint_nodes, skin_nodes[(tag, j)])
            joint_elements.append((tag, j))

            ibms.append(_bind_inv(X[j], R[j]))

            for i in range(start_idx, end_idx):
                # Mark all vertices as belonging 100% to this joint
                joints_0.append( [joint, 0., 0., 0.])
                weights_0.append([  1.0, 0., 0., 0.])


    # 4) Create the Skin referencing these joints
    #------------------------------------------------------
    skin = _create_skin(canvas, ibms, joint_nodes, skeleton_root_idx)


    # 5) Build the mesh
    #------------------------------------------------------
    if len(e.coords):
        canvas.plot_mesh(e.coords,
                        [list(reversed(face)) for face in e.triang],
                        joints_0=joints_0,
                        weights_0=weights_0,
                        skin=skin,
                        mesh_name="FrameSkinMesh",
                        node_name="FrameSkinMeshNode",
        )

    return skin_nodes, joint_nodes, joint_elements


def skin_points(model, artist, config=None, mesh=None):
    """
    Returns a dictionary mapping (point_name, j) -> glTF node index
    """
    if config is None:
        config = {
            "style": MeshStyle(color="gray"),
            "scale": 1.0,
            "outline": "",
        }
    canvas = artist.canvas 
    Ra = artist._plot_rotation

    default_points = np.array([
                    [-1.0, -1.0,  1.0],
                    [ 1.0, -1.0,  1.0],
                    [-1.0,  1.0,  1.0],
                    [ 1.0,  1.0,  1.0],
                    [ 1.0, -1.0, -1.0],
                    [-1.0, -1.0, -1.0],
                    [ 1.0,  1.0, -1.0],
                    [-1.0,  1.0, -1.0],
                ],
                dtype=canvas.float_t,
            )/10
    default_indices = np.array([
                    [0, 1, 2],
                    [3, 2, 1],
                    [1, 0, 4],
                    [5, 4, 0],
                    [3, 1, 6],
                    [4, 6, 1],
                    [2, 3, 7],
                    [6, 7, 3],
                    [0, 2, 5],
                    [7, 5, 2],
                    [5, 7, 4],
                    [6, 4, 7],
                ], dtype=canvas.index_t,
            )


    #
    # Create a skeleton root node
    #
    gltf = canvas.gltf
    skeleton_root_node = pygltflib.Node(name="PointSkeletonRoot", children=[])
    skeleton_root_idx = _append_index(gltf.nodes, skeleton_root_node)
    gltf.scenes[0].nodes.append(skeleton_root_idx)

    #
    joint_nodes = skeleton_root_node.children
    ibms = []
    skin_nodes = {}

    def _bind_inv(translation, rotmat):
        M = np.eye(4, dtype=canvas.float_t)
        return M

    #
    # 3) For each ring, create a glTF Node (joint),
    #    and assign ring vertices to that joint
    #
    I = 0
    coords = []
    triang = []
    joints_0    = []
    weights_0   = []


    for tag in model.iter_node_tags():
        # Setup mesh
        mesh = model.node_marker(None)
        if mesh is not None:
            points = mesh.points
            indices = None
            for type in mesh.cells:
                if type.type == "triangle":
                    indices = I + type.data
            if indices is None:
                raise ValueError(f"Mesh {mesh.name} has no triangle cells")
        
        else:
            points = default_points
            indices = I + default_indices 

        coords.extend(points.tolist())
        triang.extend([list(reversed(face)) for face in indices.tolist()])
        I += len(points)
        #
        X = (Ra@model.node_position(tag)).tolist()
        node = pygltflib.Node(
             translation =  X,
            #  rotation    =  Rotation.from_matrix(Ra).as_quat().tolist()
        )

        skin_nodes[tag] = _append_index(gltf.nodes, node)

        # add to skeleton root
        joint = _append_index(joint_nodes, skin_nodes[tag])

        ibms.append(_bind_inv(X, np.eye(3, dtype=canvas.float_t)))

        for i in range(len(points)):
            # Mark all vertices as belonging 100% to this joint
            joints_0.append( [joint, 0., 0., 0.])
            weights_0.append([  1.0, 0., 0., 0.])


    # 4) Create the Skin referencing these joints
    #------------------------------------------------------
    skin = _create_skin(canvas, ibms, joint_nodes, skeleton_root_idx)

    # 5) Build the mesh
    #------------------------------------------------------
    if len(coords):
        canvas.plot_mesh(coords,
                         triang,
                         joints_0=joints_0,
                         weights_0=weights_0,
                         skin=skin,
                         mesh_name="PointSkinMesh",
                         node_name="PointSkinMeshNode",
        )

    return skin_nodes

def _create_skin(canvas, ibms, joint_nodes, skeleton):
    "Create a Skin referencing given joints and add to skeleton"
    gltf = canvas.gltf

    # Flatten the inverse bind matrices into an Nx16 float32 array
    ibm_array = np.array(ibms, dtype=canvas.float_t).reshape(-1,16)

    # Create accessor to inverse bind matrices and skin
    skin = pygltflib.Skin(
        inverseBindMatrices=_append_index(gltf.accessors, pygltflib.Accessor(
            bufferView=canvas._push_data(ibm_array.tobytes(), target=None),
            componentType=GLTF_T[canvas.float_t],
            count=len(ibms),
            type="MAT4"
        )),
        joints=joint_nodes,
        skeleton=skeleton,
        name="FrameExtrusionSkin"
    )

    if not gltf.skins:
        gltf.skins = []

    return _append_index(gltf.skins, skin)



class Motion:
    """
    A helper class that accumulates multiple "states" (deformed configurations)
    and creates a time-based glTF Animation. Each call to add_state() adds
    a new keyframe at the next time step.
    """

    def __init__(self, artist=None, time_step=1.0, name="BeamDeformations"):
        """
        :param canvas:   An instance of your GltfCanvas (with .gltf).
        :param extrusion: Dict {(element_name, j): gltf_node_index, ...}
                         returned by draw_extrusions_ref().
        :param time_step: The time increment for each added state (seconds, or frames).
        :param name: The name of the final glTF animation.
        """
        self.model  = artist.model
        self.artist = artist
        self.canvas = artist.canvas

        self.time_step = time_step
        self.current_time = 0.0
        self._anim_name = name

        self._keyframes = defaultdict(lambda: {"translation": [], "rotation": []})

        self._section_skins = None

        self._point_skins = None

        self._outine_skins = None

        # Warping
        # self._mesh_morph_keyframes = []
    

    def advance(self, time=None):
        if time is None:
            self.current_time += self.time_step
        else:
            self.current_time = time

    def set_mode_state():
        pass

    def set_node_position(self, node, position, time=None):
        if time is None: 
            time = self.current_time

        self._keyframes[node]["translation"].append((time, position))

    def set_node_rotation(self, node, rotation, time=None):
        if time is None: 
            time = self.current_time

        self._keyframes[node]["rotation"].append((time, rotation))


    def set_field(self, field, time=None):
        """
        Record a keyframe for a node scale
        """
        if time is None:
            time = self.current_time

        if not hasattr(self, '_mesh_morph_keyframes'):
            self._mesh_morph_keyframes = []

        model = self.model
        field = [
            field(model.cell_nodes(element)[j]) for element,j in self._joint_elements
        ]
        self._mesh_morph_keyframes.append((time, field))

    def draw_nodes(self,
                    state=None, 
                    rotation=None, position=None):
        """
        Given a 'state' that has deformed positions and rotations for each element’s cross-section,
        record a new keyframe at the current time.

        :param state:  Some data structure that can provide displacements & rotations
                       for each (element, cross_section_index).
        """
        if self._point_skins is None:
            self._point_skins = skin_points(self.model, self.artist)
        
        skin_nodes = self._point_skins

        state = self.model.wrap_state(state, 
                                        rotation=rotation, 
                                        position=position,
                                        transform=self.artist.dofs2plot)
        model = self.model
        Ra = self.artist._plot_rotation


        for tag in model.iter_node_tags():

            # Displacements & rotations from 'state'
            key = tag
            # look up the glTF node index
            if key not in skin_nodes:
                continue

            # compute final position and rotation for cross section j
            x_def = Ra@model.node_position(tag, state=state)
            q = Rotation.from_matrix(Ra@model.node_rotation(tag,state=state)).as_quat().tolist()

            # store a keyframe
            self.set_node_position(skin_nodes[key], (x_def[0], x_def[1], x_def[2]))
            self.set_node_rotation(skin_nodes[key], q)



    def draw_sections(self,
                      state=None, rotation=None, position=None, warp=None,
                      time=None):
        """
        Given a 'state' that has deformed positions and rotations for each element’s cross-section,
        record a new keyframe at the current time.

        :param state:  Some data structure that can provide displacements & rotations
                       for each (element, cross_section_index).
        """
        if self._section_skins is None:
            self._section_skins, self._joint_nodes, self._joint_elements = \
                skin_frames(self.model, self.artist,
                            config=self.artist._config_sketch("default")["surface"]["frame"])
        
        skin_nodes = self._section_skins

        #
        # State
        #
        if state is None and rotation is None and position is None:
            return

        state = self.model.wrap_state(state, 
                                rotation=rotation, 
                                position=position,
                                transform=self.artist.dofs2plot)
        model = self.model
        Ra = self.artist._plot_rotation


        for tag in model.iter_cell_tags():
            if not model.cell_matches(tag, "frame"):
                continue

            elem = model.frame_element(tag)

            R0 = model.frame_orientation(tag).T
            # X_ref = model.cell_position(element_name)
            nes = len(list(elem.simple_samples()))

            # Displacements & rotations from 'state'
            pos_all = np.array([
                # Ra@elem.sample_position(sample, state=state) for sample in elem.simple_samples()
                Ra@model.node_position(node, state=state) 
                for node in model.cell_nodes(tag)
            ])
            rot_all = [
                Ra@Ri@R0 for Ri in state.cell_array(tag, state.rotation)
            ]

            for j in range(nes):
                # look up the glTF node index
                key = (tag, j)
                if key not in skin_nodes:
                    continue

                # compute final position and rotation for cross section j
                if False:
                    x_def = pos_all[j]
                    qx, qy, qz, qw = Rotation.from_matrix(rot_all[j]).as_quat()
                else:
                    x_def = Ra@elem.sample_position(_SimpleSample(j), state=state)
                    qx, qy, qz, qw = Rotation.from_matrix(
                        Ra@elem.sample_rotation(_SimpleSample(j), state=state)
                    ).as_quat().tolist()

                # store a keyframe
                self.set_node_position(skin_nodes[key], (x_def[0], x_def[1], x_def[2]))
                self.set_node_rotation(skin_nodes[key], (qx, qy, qz, qw))


    def add_to(self, canvas):
        """
        Build a glTF Animation from the accumulated keyframes and
        then let the canvas write the final file.
        """
        gltf = canvas.gltf
    
        if not self._keyframes:
            return

        anim = pygltflib.Animation(name=self._anim_name,
                                   samplers=[],
                                   channels=[])

        # 1) Create samplers and channels:
        #   - For each node, we have two samplers (translation, rotation)
        #   - Then two channels referencing those samplers

        # Record the sampler index for each node property as we build them
        # so we can attach channels referencing the correct sampler.
        node_position_sampler_index = {}
        node_rotation_sampler_index = {}

        # 2) Flatten and encode data for each node
        # Do them all in a single big set of buffers—time values and output values.
        # However, each node gets its own Sampler, because it has distinct times/values
        # in this implementation.
        for node_idx, track_dict in self._keyframes.items():
            pos_keyframes = track_dict["translation"]  # list of (time, (x,y,z))
            rot_keyframes = track_dict["rotation"]     # list of (time, (qx,qy,qz,qw))

            if not pos_keyframes and not rot_keyframes:
                continue


            # Sort them by time just in case user added states out of order
            pos_keyframes.sort(key=lambda x: x[0])
            rot_keyframes.sort(key=lambda x: x[0])

            # assert pos_keyframes and rot_keyframes

            if pos_keyframes:
                # Create Sampler for translation
                sampler_index_t = _append_index(anim.samplers, pygltflib.AnimationSampler(
                    input=-1,    # placeholder, fill them after creating Accessors
                    output=-1,   #
                    interpolation="LINEAR"
                ))
                node_position_sampler_index[node_idx] = sampler_index_t
                # Temporarily store the arrays so we can embed them in the glTF buffer
                # after building all samplers.
                anim.samplers[sampler_index_t].extras = {
                    "times_array": np.array([k[0] for k in pos_keyframes], dtype=canvas.float_t),
                    "vals_array":  np.array([k[1] for k in pos_keyframes], dtype=canvas.float_t)
                }
            
            if rot_keyframes:
                # Create Sampler for rotation
                sampler_index_r = _append_index(anim.samplers, pygltflib.AnimationSampler(
                    input=-1,
                    output=-1,
                    interpolation="LINEAR"
                ))
                node_rotation_sampler_index[node_idx] = sampler_index_r

                # Temporarily store the arrays so we can embed them in the glTF buffer
                # after building all samplers.
                anim.samplers[sampler_index_r].extras = {
                    "times_array": np.array([k[0] for k in rot_keyframes],   dtype=canvas.float_t),
                    "vals_array":  np.array([k[1] for k in rot_keyframes],   dtype=canvas.float_t)
                }

        # 3)
        for node_idx in self._keyframes:
            if node_idx in node_position_sampler_index:
                # Channel for translation
                anim.channels.append(pygltflib.AnimationChannel(
                    sampler=node_position_sampler_index[node_idx],
                    target=pygltflib.AnimationChannelTarget(
                        node=node_idx,
                        path="translation"
                    )
                ))

            if node_idx in node_rotation_sampler_index:
                # Channel for rotation
                anim.channels.append(pygltflib.AnimationChannel(
                    sampler=node_rotation_sampler_index[node_idx],
                    target=pygltflib.AnimationChannelTarget(
                        node=node_idx,
                        path="rotation"
                    )
                ))


        # --- Add Warp Animation Channels for Each Joint ---
        # This helper repurposes the joint node's scale (using its x-component)
        # to store the warp value.
    
        # If warp keyframes were recorded, add per-joint warp channels.
        if hasattr(self, '_mesh_morph_keyframes'):
            # self._joint_nodes should be a list of joint node indices in the same order as the warp values.
            _add_warp_animation_to_joints(anim, self._joint_nodes, self._mesh_morph_keyframes, canvas)


        # 4) Insert the time / value data into buffers.
        #    Create BufferViews and Accessors, then set up each sampler's input/output
        #    to reference the newly created accessor indices.
        for sampler in anim.samplers:
            time_accessor_idx = _append_index(gltf.accessors, pygltflib.Accessor(
                bufferView=canvas._push_data(sampler.extras["times_array"].tobytes()),
                byteOffset=0,
                componentType=FLOAT,
                count=len(sampler.extras["times_array"]),
                type="SCALAR",
                min=[float(sampler.extras["times_array"].min())],
                max=[float(sampler.extras["times_array"].max())]
            ))

            # If path=="translation" we have 3 floats, if path=="rotation" we have 4.
            # But we already know shape from sampler.extras["vals_array"].shape
            val_type = "VEC3" if sampler.extras["vals_array"].shape[1]==3 else "VEC4"

            # For morph weights, glTF expects a VEC? array; since we have one target, it is VEC1.
            if sampler.extras["vals_array"].shape[1] == 1:
                val_type = "SCALAR"

            vals_accessor_idx = _append_index(gltf.accessors, pygltflib.Accessor(
                bufferView=canvas._push_data(sampler.extras["vals_array"].tobytes()),
                byteOffset=0,
                componentType=FLOAT,
                count=len(sampler.extras["vals_array"]),
                type=val_type
            ))
            sampler.input  = time_accessor_idx
            sampler.output = vals_accessor_idx

            # Remove extras so we dont JSON-serialize large arrays
            del sampler.extras["times_array"]
            del sampler.extras["vals_array"]

        # 5) Attach this new Animation to the glTF
        if not gltf.animations:
            gltf.animations = []

        _append_index(gltf.animations, anim)


def _add_warp_animation_to_joints(anim, joint_nodes, warp_keyframes, canvas):
    # Sort warp keyframes by time.
    warp_keyframes.sort(key=lambda x: x[0])
    num_joints = len(joint_nodes)

    # Build per-joint keyframe data.
    joint_keyframes = [[] for _ in range(num_joints)]
    for t, warp_list in warp_keyframes:
        # Each warp_list must have num_joints values.
        for j in range(num_joints):
            joint_keyframes[j].append((t, warp_list[j]))

    # For each joint, create an animation channel targeting its scale.
    for j, node_index in enumerate(joint_nodes):
        times = np.array([t for t, _ in joint_keyframes[j]], dtype=canvas.float_t)
        # Pack the warp value in the x-component; keep y and z at 1.
        scale_values = np.array([[w, 1.0, 1.0] for _, w in joint_keyframes[j]], dtype=canvas.float_t)
        sampler_index = _append_index(anim.samplers, pygltflib.AnimationSampler(
            input=-1,
            output=-1,
            interpolation="LINEAR"
        ))
        anim.samplers[sampler_index].extras = {
            "times_array": times,
            "vals_array": scale_values
        }
        anim.channels.append(pygltflib.AnimationChannel(
            sampler=sampler_index,
            target=pygltflib.AnimationChannelTarget(
                node=node_index,
                path="scale"
            )
        ))



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio M. Perez
#
import sys
import warnings
import numpy as np
from scipy.spatial.transform import Rotation

import shps.curve

import veux
from veux.model import read_model
from veux.errors import RenderError
from veux.config import MeshStyle
from shps.frame.extrude import FrameMesh
from dataclasses import dataclass

@dataclass
class ExtrusionCollection:
    triang: list
    coords: list
    caps: list
    no_outline: set
    in_outline: set

def add_extrusion(extr, e, x, R, I, caps=None):

    ring_ranges = extr.ring_ranges()

    p = extr.vertices()
    indices = extr.triangles()

    if len(indices) == 0 or len(ring_ranges) == 0:
        return 0

    e.triang.extend([I + T for T in indices])

    for (j, start_idx, end_idx) in ring_ranges:
        for i in range(start_idx, end_idx):
            e.coords.append(x[j] + R[j] @ p[i])

    if caps:
        nen = len(x)
        noe = ring_ranges[0][-1]
        caps[0].append(            I+np.arange(noe))
        caps[1].append((nen-1)*noe+I+np.arange(noe))

    return len(p) #len(indices)


def draw_extrusions3(model, canvas, state=None, config=None, Ra=None):
    if config is None:
        config = {"style": MeshStyle(color="gray")}
    if Ra is None:
        Ra = np.eye(3)

    scale = config.get("scale", 1.0)

    # 1) Build local geometry
    I = 0
    caps = []
    e = ExtrusionCollection([], [], [], set(), set())
    for tag in model.iter_cell_tags():
        if not model.cell_matches(tag, "frame") and not model.cell_matches(tag, "truss"):
            continue

        R0 = model.frame_orientation(tag)
        if R0 is None:
            warnings.warn(f"Frame {tag} has no orientation")
            continue
        else:
            R0 = R0.T

        X_ref = np.array([
            Ra@model.node_position(node) for node in model.cell_nodes(tag)
        ])
        nen = len(X_ref)

        if state is not None:
            x = np.array([
                Ra@model.node_position(node, state=state) for node in model.cell_nodes(tag)
            ])
            # u = state.cell_array(tag, state.position)
            # x = shps.curve.displace(X_ref, u, nen)
            R = [Ra@Ri@R0 for Ri in state.cell_array(tag, state.rotation)]
        else:
            x = X_ref
            R = [Ra@R0 for _ in range(nen)]

        sections = [model.frame_section(tag, i) for i in range(len(x))]
        if sections[0] is None or sections[-1] is None:
            continue

        icap, jcap = [], []
        #
        # Exterior
        #
        extr = FrameMesh(len(x),
                        [s.exterior() for s in sections],
                        scale=scale,
                        do_end_caps=False)

        ne = add_extrusion(extr, e, x, R, I, [icap, jcap])

        si = sections[0]
        if len(si.exterior()) > 35:
            for i in range(ne):
                e.no_outline.add(I+i)

        #
        # Interior
        #
        ni = 0
        for i in range(len(si.interior())):
            if si.interior()[i] is None or len(si.interior()[i]) == 0:
                continue

            extr = FrameMesh(len(x),
                    [s.interior()[i] for s in sections],
                    scale=scale,
                    do_end_caps=False)
        
            nij = add_extrusion(extr, e, x, R, I+ne+ni, [icap, jcap])
            for i in range(nij):
                # e.no_outline.add(I+ne+ni+i)
                e.in_outline.add(I+ne+ni+i)
            ni += nij

        I += ni + ne

        #
        # Caps
        #
        try:
            face = si.triangles()

        except Exception as ex:
            warnings.warn(f"Earcut failed with message: {ex}")
            continue

        iicap = [ i for j in icap for i in j ]
        ijcap = [ i for j in jcap for i in j ]
        caps.extend([
            [iicap[i] for i in face],
            [ijcap[i] for i in face]
        ])


    # Draw mesh
    if e.coords:
        mesh = canvas.plot_mesh(e.coords,
                     [list(reversed(face)) for face in e.triang],
                     style=config["style"])

    # Draw caps
    if len(caps) > 0:
        for cap in caps:
            try:
                canvas.plot_mesh(mesh.vertices, cap, style=config["style"])
            except Exception as ex:
                print(ex, file=sys.stderr)

    # Draw outlines
    if "outline" not in config:
        return

    triang = e.triang
    nan = np.array([0,0,0], dtype=float)*np.nan
    IDX = np.array(((0,2),(0,1)))
    coords = np.array(e.coords)
    try:
        if True: #"tran" in config["outline"]:
            tri_points = np.array([
                coords[idx]  if (j+1)%3 else nan
                for j,idx in enumerate(np.array(triang).reshape(-1))
            ])
            if len(tri_points):
                canvas.plot_lines(tri_points,
                                  style=config["line_style"])
        if "long" in config["outline"]:
            tri_points = np.array([
                coords[i]  if j%2 else nan
                for j,idx in enumerate(np.array(triang)) for i in idx[IDX[j%2]] if j not in e.no_outline
            ])

            if len(tri_points):
                canvas.plot_lines(tri_points,
                            style=config["line_style"])
    except Exception as ex:
        warnings.warn(f"Failed to draw outline with message: {ex}")
        return

class so3:
    @classmethod
    def exp(cls, vect):
        return Rotation.from_rotvec(vect).as_matrix()

def _add_moment(artist, loc, axis):
    import meshio
    mesh_data = meshio.read(veux.assets/'chrystals_moment.stl')
    coords = mesh_data.points

    coords = np.einsum('ik, kj -> ij',  coords,
                       so3.exp([0, 0, -np.pi/4])@so3.exp(axis))
    coords = 1e-3*coords + loc
    for i in mesh_data.cells:
        if i.type == "triangle":
            triangles =  i.data #mesh_data.cells['triangle']
            break

    artist.canvas.plot_mesh(coords, triangles)


def _render(sam_file, res_file=None, **opts):
    # Configuration is determined by successively layering
    # from sources with the following priorities:
    #      defaults < file configs < kwds 

    config = veux.config.Config()


    if sam_file is None:
        raise RenderError("Expected positional argument <sam-file>")

    # Read and clean model
    if not isinstance(sam_file, dict):
        model = read_model(sam_file)
    else:
        model = sam_file

    if "RendererConfiguration" in model:
        veux.apply_config(model["RendererConfiguration"], config)

    veux.apply_config(opts, config)

    artist = veux.FrameArtist(model, **config)

    draw_extrusions3(artist.model, artist.canvas, config=opts)

    # -----------------------------------------------------------

    soln = veux.model.read_state(res_file, artist.model, **opts)
    if soln is not None:
        if "time" not in opts:
            soln = soln[soln.times[-1]]

        draw_extrusions3(artist.model, artist.canvas, soln, opts)
        # -----------------------------------------------------------
        _add_moment(artist,
                    loc  = [1.0, 0.0, 0.0],
                    axis = [0, np.pi/2, 0])
        # -----------------------------------------------------------

    artist.draw()
    return artist


if __name__ == "__main__":
    import veux.parser
    config = veux.parser.parse_args(sys.argv)

    try:
        artist = _render(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])


        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            import veux.server
            server = veux.server.Server(glb=artist.canvas.to_glb(),
                                        viewer=config["viewer_config"].get("name", None))
            server.run(config["server_config"].get("port", None))

        elif hasattr(artist.canvas, "to_html"):
            import veux.server
            server = veux.server.Server(html=artist.canvas.to_html())
            server.run(config["server_config"].get("port", None))

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(e, file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=sys.argv[0]), file=sys.stderr)
        sys.exit()




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Copyright (c) 2025, Claudio M. Perez
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#
from collections import defaultdict
from veux.state  import StateSeries, BasicState, GroupSeriesSE3, GroupStateSE3, GroupStateSO3, Rotation
from xsection.library import Rectangle
import warnings

import numpy as np

from pathlib import Path
from urllib.parse import urlparse
try:
    import orjson as json
except ImportError:
    import json

from veux.frame._section import SectionGeometry

class Model:
    def __iter__(self):
        # this method allows: nodes, cells = Model(mesh)
        return iter((self.nodes, self.cells))

    def make_state(self, state, **kwds):
        # This just needs to return an opaque object that can be passed into the methods below
        return state 
    
    def node_marker(self, tag):
        return self._node_marker

    def node_position(self, tag, **kwds): ...

    def node_information(self, tag): ...

    def iter_cells_tags(self, filt=None): ...

    def cell_matches(self, tag, type): ...

    def cell_type(self, tag):       ... # line triangle quadrilateral 

    def cell_exterior(self, tag):   ...

    def cell_interior(self, tag):   ...

    def cell_quadrature(self, tag):  return []

    def cell_prototypes(self):  return []

    def cell_rotation(self, tag, state): raise NotImplementedError

    def cell_position(self, tag, state): raise NotImplementedError

    def cell_outline(self,  tag):   ...

    def cell_information(self, tag): ...

    def cell_triangles(self, tag):  ...


# Constants
_EYE3 = np.eye(3)

_OUTLINES = {
    None:      None,
    "square":  np.array([[-1., -1.],
                         [ 1., -1.],
                         [ 1.,  1.],
                         [-1.,  1.]])/4.0,

    "tee":     np.array([[ 6.0,  0.0],
                         [ 6.0,  4.0],
                         [-6.0,  4.0],
                         [-6.0,  0.0],
                         [-2.0,  0.0],
                         [-2.0, -8.0],
                         [ 2.0, -8.0],
                         [ 2.0,  0.0]])/10
}


def _is_frame(el):
    name = el["type"].lower()
    return     "beam"  in name \
            or "dfrm"  in name \
            or "frame" in name

def _is_truss(el):
    name = el["type"].lower()
    return "truss" in name #or "twonodelink" in name


def _is_plane(el):
    name = el["type"].lower()
    return "quad" in name or "shell" in name or "tri" in name

def _is_solid(el):
    name = el["type"].lower()
    return "brick" in name or "tetra" in name


def _orient_frame(xi, xj, angle):
    """
    Calculate the coordinate transformation vector.
    xi is the location of node I, xj node J,
    and `angle` is the rotation about the local axis in degrees

    By default local axis 2 is always in the 1-Z plane, except if the object
    is vertical and then it is parallel to the global X axis.
    The definition of the local axes follows the right-hand rule.
    """

    # The local 1 axis points from node I to node J
    dx, dy, dz = e1 = xj - xi
    # Global z
    E3 = np.array([0, 0, 1])

    # In Sap2000, if the element is vertical, the local y-axis is the same as the
    # global x-axis, and the local z-axis can be obtained by cross-multiplying
    # the local x-axis with the local y-axis.
    if dx == 0 and dy == 0:
        e2 = np.array([1, 0, 0])

    # Otherwise, the plane composed of the local x-axis and the local
    # y-axis is a vertical plane. In this
    # case, the local z-axis can be obtained by the cross product of the local
    # x-axis and the global z-axis.
    else:
        e2 = np.cross(E3, e1)

    e3 = np.cross(e1, e2)

    # Rotate the local axis using the Rodrigue rotation formula
    # convert from degrees to radians
    angle = angle / 180 * np.pi
    e3r = e3 * np.cos(angle) + np.cross(e1, e3) * np.sin(angle)
    # Finally, the normalized local z-axis is returned
    return e3r / np.linalg.norm(e3r)


def read_model(filename:str, shift=None, verbose=False)->dict:
    if isinstance(filename, str) and filename.endswith(".tcl"):
        import opensees.tcl
        try:
            with open(filename, "r") as f:
                interp = opensees.tcl.exec(f.read(), silent=True, analysis=False)
        except UnicodeDecodeError:
            with open(filename, "r", encoding="latin1") as f:
                interp = opensees.tcl.exec(f.read(), silent=True, analysis=False)
        return interp.serialize()

    elif isinstance(filename, str) and (
        filename.endswith(".s2k") or filename.endswith(".$2k") or filename.endswith(".$br")) or filename.endswith(".b2k"):
        from xcsi import Job
        model = Job(filename, mode="visualize").instance().model
        return model.asdict()

    elif isinstance(filename, str) and filename.endswith(".inp"):
        from xcae import Job
        model = Job(filename, mode="visualize").instance().model
        return model.asdict()

    elif isinstance(filename, str) and filename.endswith(".vtk"):
        import meshio
        from .plane import PlaneModel
        return PlaneModel(meshio.read(filename))

    elif isinstance(filename, str) and filename.endswith(".inp"):
        pass

    try:
        with open(filename,"r") as f:
            sam = json.loads(f.read())

    except TypeError:
        sam = json.loads(filename.read())

    return sam


def read_state(res_file,
               model=None,
               rotation=None,
               position=None,
               time=None,
               scale=None,
               transform=None,
               recover_rotations=None,
               **opts):
    """
    """

    # Turn res_file into res
    if hasattr(res_file, "read"):
        import yaml
        res = yaml.load(res_file, Loader=yaml.Loader)

    elif isinstance(res_file, (str,Path)):
        res_path = urlparse(res_file)
        if "json" in res_path[2]:
            with open(res_path[2], "r") as f:
                res = json.loads(f.read())
        else:
            with open(res_path[2], "r") as f:
                res = yaml.load(f, Loader=yaml.Loader)

        if res_path[4]: # query parameters passed
            res = res[int(res_path[4].split("=")[-1])]
    else:
        res = res_file


    #
    # Create the object
    #
    if model.ndm == model.ndf:
        if isinstance(res, np.ndarray) and len(res.shape) > 1:
            vectors = {
                i: res[i] for i in range(len(res))
            }

        elif isinstance(res, np.ndarray) and res.shape[0] == len(model.iter_node_tags()):
            vectors = {tag: res[i]  for i, tag in enumerate(model.iter_node_tags())}

        elif isinstance(res, np.ndarray):
            # TODO: need to get node-dof mapping
            raise NotImplementedError

        elif callable(res):
            vectors = {tag: res(tag) for tag in model.iter_node_tags()}

        else:
            vectors = res

        return BasicState(vectors, model, transform=transform, scale=scale)


    else:
        # Create a GroupStateSE<n>
        xdof = slice(0,model.ndm)
        rdof = 2 if model.ndm == 2 else slice(3, 6)
        if transform is None:
            transform = np.eye(6)

        if isinstance(res, np.ndarray):
            raise NotImplementedError

        elif callable(res):
            # eg, state=model.nodeDisp
            position_state = BasicState(
                {k: transform[:3,:model.ndm]@res(k)[xdof] for k in model.iter_node_tags()},
                model,
                time=time,
                scale=scale
            )
            if model.ndm == 2:
                rotation_state = GroupStateSO3({
                        k: Rotation.from_rotvec([0, 0, res(k)[rdof]]) 
                        for k in model.iter_node_tags()
                    }, 
                    model, 
                    # scale=scale, 
                    transform=transform[3:,3:], 
                    time=time)
            else:
                rotation_state = GroupStateSO3({
                        k: Rotation.from_rotvec(res(k)[rdof]) 
                        for k in model.iter_node_tags()
                    }, 
                    model, 
                    # scale=scale, 
                    transform=transform[3:,3:], 
                    time=time
                )

        # FEDEAS
        elif isinstance(res, dict) and ("IterationHistory" in res or "ConvergedHistory" in res):

            history = StateSeries(res, model,
                        transform =transform,
                        recover_rotations=recover_rotations
                    )

            if recover_rotations is not None:
                history = GroupSeriesSE3(history, model, recover_rotations=recover_rotations, transform=transform)

            if time is not None:
                return history[time]
            else:
                # TODO: This function should never return a series;
                # implement create_series for that
                return history

        # res is a Dict from node tag to nodal values
        elif isinstance(res, dict):
            position_state = BasicState(
                {k: transform[:3,:model.ndm]@v[xdof] for k,v in res.items()}, 
                model, 
                time=time,
                scale=scale
            )
            if model.ndm == 2:
                rotation_state = GroupStateSO3({
                        k: Rotation.from_rotvec([0, 0, v[rdof]]) if len(v[rdof]) == 1 else Rotation.from_quat(v[rdof])
                        for k,v in res.items()
                    },
                    model, 
                    scale=scale, 
                    transform=transform[3:,3:], 
                    time=time
                )
            else:
                rotation_state = GroupStateSO3(
                        {
                            k: Rotation.from_rotvec(v[rdof])
                            for k,v in res.items()
                        },
                        model,
                      # scale=scale,
                        transform=transform[3:,3:],
                        time=time
                    )

        if callable(position):
            position_state = BasicState(
                {k: transform[:3,:model.ndm]@position(k)[xdof] for k in model.iter_node_tags()},
                model, 
                time=time,
                scale=scale
            )

        if callable(rotation) and model.ndm==2:
            rotation_state = GroupStateSO3(
                {k: Rotation.from_rotvec([0,0,rotation(k)]) for k in model.iter_node_tags()},
                model,
                transform=transform[3:,3:],
                time=time,
                # scale=scale
            )

        elif callable(rotation) and model.ndm==3:
            rotation_state = GroupStateSO3(
                {k: Rotation.from_quat(rotation(k)) for k in model.iter_node_tags()},
                model, 
                transform=transform[3:,3:],
                time=time,
                # scale=scale
            )

        return GroupStateSE3((position_state, rotation_state), model)



class FrameModel:

    def __init__(self,
                 sam:dict,
                 shift = None,
                 frame_outlines=None,
                 frame_shape = None,
                 frame_samples=None,
                 node_marker=None,
                 node_marker_scale=1.0,
                 xmodel=None,
                 **kwds):

        # number of dimensions of artist, not model?
        nda = 3
        R = np.eye(nda) #if rot is None else rot

        if shift is None:
            shift = np.zeros(nda)
        else:
            shift = np.asarray(shift)


        self._xmodel = xmodel
        self._node_marker = node_marker
        if node_marker_scale is not None and node_marker is not None:
            self._node_marker.points *= node_marker_scale
        #
        self._data = _from_opensees(sam, shift, R)# output

        self.ndm = self._data["ndm"]
        self.ndf = self._data["ndf"]

        #
        # Section data
        #
        from .frame._section import SectionGeometry

        # 1) Supplementary data
        self._frame_sections = {}
        self._frame_samples = frame_samples
        for name,section in self._data["sections"].items():
            _add_section_shape(section, self._data["sections"], self._frame_sections, self.ndm)
        
        self._frame_sections = {
            k: SectionGeometry(v) for k,v in self._frame_sections.items() if len(v) > 0
        }
        

        self._frame_outlines = {}
        if isinstance(frame_outlines, dict):

            for key, polygons in frame_outlines.items():
                if hasattr(polygons, "exterior"):
                    self._frame_outlines[key] = [polygons]
                elif isinstance(polygons, list) and all(hasattr(i, "exterior") for i in polygons):
                    self._frame_outlines[key] = polygons
                elif hasattr(polygons[0][0], "__len__"):
                    # list of numpy arrays
                    self._frame_outlines[key] = [SectionGeometry(i) for i in polygons]
                else:
                    self._frame_outlines[key] = [SectionGeometry(polygons)]

        elif frame_outlines is None and "extrude_outline" not in kwds:
            # TODO: Make this dict of list of sections
            self._frame_outlines = _get_frame_outlines(self)

        # 2) Default extrusion
        self._extrude_default = SectionGeometry(_OUTLINES[kwds.get("extrude_default", "square")])

        # 3) Forced extrusion
        if "extrude_outline" in kwds and frame_shape is None:
            frame_shape = kwds["extrude_outline"]
        if frame_shape is not None:
            if hasattr(frame_shape, "exterior"):
                self._frame_shape = SectionGeometry(
                                          exterior=frame_shape.exterior(),
                                          interior=frame_shape.interior(),
                                          warping=kwds.get("section_warping", None))

            elif isinstance(kwds["extrude_outline"], str):
                self._frame_shape = SectionGeometry(_OUTLINES[kwds["extrude_outline"]])
            else:
                raise ValueError("extrude_outline must be a SectionGeometry or a string")
        else:
            self._frame_shape = None

        self._extrude_scale   = kwds.get("extrude_scale",   1.0)


    def __getitem__(self, key):
        # TODO: Remove this method
        return self._data[key]
    
    def node_marker(self, tag):
        return self._node_marker

    def wrap_state(self,
                    state=None,
                    position=None,
                    rotation=None,
                    scale=None, 
                    transform=None, 
                    recover_rotations=None,
                    **kwds):
        """
        Parameters
        ==========
        state : dict, np.ndarray, callable, optional
            The state of the model, see :ref:`State`. Default is None, in which case the reference state of the model is rendered.
        rotation : np.ndarray, optional
            A callable that returns a quaternion representing the rotation of a given node. In OpenSeesRT models, the `nodeRotation <https://xara.so/user/manual/output/nodeRotation.html>`_ method of a ``Model`` object is typically used.
        """
        if position is not None or rotation is not None or isinstance(state, (dict, np.ndarray)) or callable(state):
            if self.ndm == self.ndf:
                # TODO: This should be gurranteed to return a BasicState;
                # perhaps call the one in plane/__init__.py
                return read_state(state,
                                model=self,
                                scale=scale,
                                transform=transform)
            else:
                # TODO: This should be gurranteed to return a GroupStateSEn
                return read_state(
                                state,
                                position=position,
                                rotation=rotation,
                                model=self,
                                scale=scale,
                                transform=transform,
                                recover_rotations=recover_rotations)
        else:
            return state

    def cell_nodes(self, tag=None):
        if tag is None:
            if not hasattr(self, "_cell_nodes"):
                self._cell_nodes = {k: e["nodes"] for k, e in self["assembly"].items()}
            return self._cell_nodes
        else:
            return self["assembly"][tag]["nodes"]

    def cell_indices(self, tag=None):
        if not hasattr(self, "_cell_indices"):
            self._cell_indices = {
                elem["name"]: tuple(self.node_indices(n) for n in elem["nodes"])
                for elem in self["assembly"].values()
            }

        if tag is not None:
            return self._cell_indices[tag]
        else:
            return self._cell_indices

    def cell_properties(self, tag=None):
        if tag is not None:
            return self["assembly"][tag]

    def cell_prototypes(self)->"iter":
        exclude_keys = {"type", "instances", # "nodes",
                        "crd", "crdTransformation"}

        if not self["prototypes"]:
            elem_types = defaultdict(dict)

            for elem in self["assembly"].values():
                if not self.cell_matches(elem["name"], "frame"):
                    continue
                type = elem["type"]
                if type not in elem_types:
                    elem_types[type]["name"] = type
                    elem_types[type]["variants"] = []
                    elem_types[type]["instances"] = [elem["name"]]
                    elem_types[type]["properties"] = {
                        k: v for k,v in elem.items() if k not in exclude_keys
                    }
                else:
                    elem_types[type]["instances"].append(elem["name"])


            elem_types = list(elem_types.values())
        else:
            elem_types = [
                {
                    "name": f"{elem['type']}<{elem['name']}>",
                    "instances": [self["assembly"][i]["name"] for i in elem["instances"]],
                    "properties":  [
                        [str(v) for k,v in elem.items() if k not in exclude_keys]
                        #for _ in range(len(elem["instances"]))
                    ]*(len(elem["instances"])),
                    "coords": [self["assembly"][i]["crd"] for i in elem["instances"]],
                    "keys":   [k for k in elem.keys() if k not in exclude_keys]

                } for elem in self["prototypes"].get("elements", [])
            ]
        return elem_types

    def iter_node_tags(self):
        for tag in self["nodes"]:
            yield tag

    def iter_cell_tags(self):
        for tag in self["assembly"]:
            yield tag

    def node_properties(self, tag=None)->dict:
        return self["nodes"][tag]

    def node_indices(self, tag=None):
        if not hasattr(self, "_node_indices"):
            self._node_indices = {
                tag: i for i, tag in enumerate(self["nodes"])
            }
        return self._node_indices[tag]

    def node_rotation(self, tag=None, state=None):
        if self.ndm == self.ndf:
            return _EYE3

        if state is None:
            eye = np.eye(3)
            if tag is None:
                return [eye for i in self.iter_node_tags()]
            else:
                return eye

        else:
            return state.node_array(tag, dof=state.rotation)

    def node_position(self, tag=None, state=None):

        if tag is None:
            pos = np.array([n["crd"] for n in self["nodes"].values()])
        else:
            pos = self["nodes"][tag]["crd"]

        if state is not None:
            pos = pos + state.node_array(tag, dof=state.position)

        return pos

    def cell_matches(self, tag, type=None)->bool:
        elem = self["assembly"][tag]

        if type == "prism":
            return _is_frame(elem) and (
                "prism" in elem["type"].lower()
                or "elastic" in elem["type"].lower()
            )
        if type == "frame":
            return _is_frame(elem)
        
        if type == "truss":
            return _is_truss(elem)

        if type == "plane":
            return _is_plane(elem)
        
        if type == "solid":
            return _is_solid(elem) 

        return False

    def cell_position(self, tag, state=None):
        if isinstance(tag, list):
            return np.array([[
                    self.node_position(node, state) for node in self["assembly"][t]["nodes"]
                ] for t in tag
            ])
        else:
            return np.array([ self.node_position(node, state)
                              for node in self["assembly"][tag]["nodes"] ])

    def frame_element(self, tag):
        """
        Iterate over all frame elements in the model.
        """
        from veux.frame._element import _FrameElement
        from veux.frame._section import SectionGeometry
        if self.cell_matches(tag, "truss"):
            return _FrameElement(
                tag=tag,
                vmodel=self,
                xmodel=self._xmodel,
                samples=2
            )

        if self.cell_matches(tag, "frame"):
            frame_shape = self._frame_shape

            if frame_shape is None and self.ndm == 2:
                frame_shape = SectionGeometry([[-1,0],[1,0]])
            

            return _FrameElement(tag=tag,
                                vmodel=self,
                                xmodel=self._xmodel,
                                samples=self._frame_samples,
                                section_override=frame_shape
            )

    def cell_exterior(self, tag):
        """
        return an array of node indices that outline the cell's boundary.
        This is related to finding an Eulerian path through the element's
        connectivity graph.
        """
        type = self["assembly"][tag]["type"].lower()

        if "frm" in type or "beamcol" in type:
            return self.cell_indices(tag)
        
        elif self.cell_matches(tag, "truss"):
            return self.cell_indices(tag)

        elif ("quad" in type or \
              "shell" in type and ("q" in type) or ("mitc" in type)):
            return self.cell_indices(tag)[:4]


        elif ("tri" in type or \
              "shell" in type and ("t" in type)):
            return self.cell_indices(tag)

        elif "tetra" in type:
            indices = self.cell_indices(tag)
            return [indices[i] for i in [0, 1, 2, 3, 0, 2, 3, 1]]

        elif "brick" in type or "hex" in type:
            i = self.cell_indices(tag)
            if len(i) == 8:
                return [
                    i[j] for j in (0, 1, 2, 3, 0, 4, 5, 6, 7, 4, 5, 1, 2, 6, 7, 3)
                ]
            else:
                # TODO: Currently not handling higher-order bricks
                return []

        return []

    def cell_quadrature(self, tag):
        if self.cell_matches(tag, "frame"):
            if self.cell_matches(tag, "prism"):
                return []
            rule = self["assembly"][tag].get("integration", None)
            if rule is None:
                return []
            
            n = len(self["assembly"][tag]["sections"])
            from shps.gauss import iquad
            for xi, wt in zip(*iquad(n=n, rule=rule["type"].lower(), bounds=(0, 1))):
                yield xi, wt

    def cell_interpolation(self, tag):
        pass

    def cell_triangles(self, tag):
        """
        """
        type = self["assembly"][tag]["type"].lower()

        if self.cell_matches(tag, "frame"):
            return []

        elif ("tri" in type or
             ("shell" in type and ("dkgt" in type))):
            return [self.cell_indices(tag)]

        elif ("quad" in type or
             ("shell" in type and ("q" in type) or ("mitc" in type))):
            nodes = self.cell_indices(tag)

            if len(nodes) == 3:
                return nodes

            if len(nodes) in {4,8,9}:
                return [[nodes[0], nodes[1], nodes[2]],
                        [nodes[2], nodes[3], nodes[0]]]

        elif "tetra" in type:
            nodes = self.cell_indices(tag)
            return [[nodes[0], nodes[2], nodes[1]],
                    [nodes[0], nodes[1], nodes[3]],
                    [nodes[0], nodes[3], nodes[2]],
                    [nodes[1], nodes[2], nodes[3]]]

        elif "brick" in type:
            nodes = self.cell_indices(tag)

            if len(nodes) == 8:
                triangles = []
                for face in ((0, 3, 2, 1), (0, 1, 5, 4), (0, 4, 7, 3),
                             (6, 7, 4, 5), (6, 2, 3, 7), (6, 5, 1, 2)):
                    triangles.extend([
                            [nodes[face[0]], nodes[face[1]], nodes[face[2]]],
                            [nodes[face[2]], nodes[face[3]], nodes[face[0]]]
                    ])
                return triangles

        return []

    def _frame_section(self, tag):
        if tag is None:
            return self._extrude_default
        return self._frame_sections[int(tag)] if int(tag) in self._frame_sections else self._extrude_default

    def frame_section(self, tag, coord=None)->"SectionGeometry":
        from .frame._section import SectionGeometry

        if not self.cell_matches(tag, "frame"):
            return None


        # # Initialize frame outlines
        # if self._frame_outlines is None:
        #     # TODO: Make this dict of list of sections
        #     self._frame_outlines = _get_frame_outlines(self)

        # sections = []
        # if self._extrude_outline is not None:
        #     sections = [
        #         SectionGeometry(self._extrude_outline.exterior()*self._section_area(tag, 0))
        #     ]*2 #*self._extrude_scale

        # elif tag in self._frame_outlines:
        #     sections = self._frame_outlines[tag]

        # elif self._extrude_default is not None:
        #     sections = [self._extrude_default]*2 #*self._extrude_scale
        elem = self.frame_element(tag)
        sections = [elem.sample_section(i) for i in elem.simple_samples()]


        # We have sections[], now get the section at the given coordinate
        if len(sections) == 0:
            # print(f"Empty sections for {tag}", file=sys.stderr)
            return

        elif len(sections) == 1:
            return sections[0]

        # Interpolate coord
        elif len(sections) >= 2:
            def interpolate(values, x):
                n = len(values) - 1
                idx = np.clip(int(x * n), 0, n - 1)  # Find the lower bound index
                t = x * n - idx  # Fractional part for interpolation
                return (1 - t) * values[idx] + t * values[idx + 1]

            if coord is None:
                coord = 0.5

            exterior = np.array([s.exterior() for s in sections])
            interior = np.array([s.interior() for s in sections])
            return SectionGeometry(
                interpolate(exterior, coord),
                interpolate(interior, coord)
            )


    def frame_orientation(self, tag, state=None):
        el  = self["assembly"][tag]
        xyz = el["crd"]

        v1  = xyz[-1] - xyz[0]
        L   = np.linalg.norm(v1)
        e1  = v1/L

        if self.ndm == 2:
            v2 =  -np.cross(e1, np.array([0, 0, 1]))
            # v2 =   np.cross(e1, np.array([0, 1, 0]))


        if self.cell_matches(tag, "truss"):
            v3 = _orient_frame(el["crd"][0], el["crd"][1], 0)
            v2 = -np.cross(e1, v3)

        elif "yvec" in el["trsfm"] and el["trsfm"]["yvec"] is not None:
            v2  = np.array(el["trsfm"]["yvec"])

        elif "vecxz" in el["trsfm"]:
            v13 =  np.atleast_1d(el["trsfm"]["vecxz"])
            v2  = -np.cross(e1,v13)

        else:
            v3 = _orient_frame(el["crd"][0], el["crd"][1], 0)
            v2  = -np.cross(e1,v3)

        e2 = v2 / np.linalg.norm(v2)
        v3 = np.cross(e1,e2)
        e3 = v3 / np.linalg.norm(v3)
        return np.stack([e1,e2,e3])


class FiberModel(Model):
    ndm = 2 
    ndf = 1
    def __init__(self, fibers, patches=None):
        self._fibers = fibers

    def iter_cell_tags(self):
        yield 1

    def cell_exterior(self):
        return 
        yield 

    def cell_interior(self):
        return 
        yield

    def wrap_state(self, state, **kwds):
        return state

    def cell_quadrature(self, tag):
        if tag == 1:
            for fiber in self._fibers:
                if "location"  in fiber:
                    x = fiber["location"]
                elif "coord" in fiber:
                    x = fiber["coord"]
                else:
                    x = fiber["y"], fiber["z"]
                yield (*x, 0), fiber["area"]


def _from_opensees(sam: dict, shift, R):
    # Process OpenSees JSON format

    # TODO?
    R = np.eye(3)

    try:
        sam = sam["StructuralAnalysisModel"]
    except KeyError:
        pass

    geom = sam.get("geometry", sam.get("assembly"))

#   ndm = len(next(iter(nodes.values()))["crd"])

    if len(geom["nodes"]) > 0:
        try:
            #coord = np.array([R@n.pop("crd") for n in geom["nodes"]], dtype=float) + shift
            coord = np.array([R@n["crd"] for n in geom["nodes"]], dtype=float) + R@shift
            ndm = 3
        except:
            coord = np.array([R@[*n["crd"], 0.0] for n in geom["nodes"]], dtype=float) + shift
            ndm = 2
    else:
        ndm = 3

    nodes = {
        n["name"]: {**n, "crd": coord[i], "idx": i}
            for i,n in enumerate(geom["nodes"])
    }

    try:
        ndf = next(iter(nodes.values())).get("ndf", None)
    except StopIteration:
        # No nodes
        ndf = 6

    trsfm = {}
    for t in sam.get("properties", {}).get("crdTransformations", []):
        trsfm[int(t["name"])] = {
            k: val for k,val in t.items() if k not in {"vecxz", "vecInLocXZPlane"}
        }
        if ndm == 3:
            trsfm[int(t["name"])]["vecxz"] = R@(t.get("vecxz", None) or t["vecInLocXZPlane"])

    def _make_transform(e):
        if "transform" in e and int(e["transform"]) in trsfm:
            return trsfm[int(e["transform"])]

        if "crdTransformation" in e and int(e["crdTransformation"]) in trsfm:
            return trsfm[int(e["crdTransformation"])]

        if "yvec" in e:
            return dict(yvec=R@e["yvec"])

    elems =  {
        e["name"]: dict(
            **e,
            crd=np.array([nodes[n]["crd"] for n in e["nodes"]], dtype=float),
            trsfm=_make_transform(e)
        ) for e in geom["elements"]
    }

    try:
        sections = {s["name"]: s for s in sam["properties"]["sections"]}
    except:
        sections = {}

    output = dict(nodes=nodes,
                  assembly=elems,
                  sam=sam,
                  sections=sections,
                  prototypes=sam.get("prototypes", {}),
                  ndm=ndm,
                  ndf=ndf
    )

    return output


def collect_outlines(model):
    return _get_frame_outlines(_from_opensees(model, [0, 0, 0], np.eye(3)))


def _add_section_shape(section, sections, outlines, ndm):

    tag = int(section["name"])
    if "section" in section:
        # Treat aggregated sections
        child_tag = int(section["section"])
        if child_tag not in outlines:
            _add_section_shape(sections[section["section"]], sections, outlines, ndm)

        outlines[tag] = outlines[child_tag]

    elif "bounding_polygon" in section:
        # Rotation to change coordinates from x-y to z-y
        R = np.array(((0,-1),
                      (1, 0))).T
        outlines[tag] = [R@s for s in section["bounding_polygon"]]

    elif "fibers" in section and ndm > 2:
        points = np.array([
            f.get("coord", None) or f["location"] for f in section["fibers"]
        ])

        if len(points) == 1:
            A = sum(i["area"] for i in section["fibers"])

            if A is None:
                return

            b = d = np.sqrt(A)
            outlines[tag] = Rectangle(b, d).exterior()
            return

        elif len(points) > 2:
            try:
                from veux.utility.alpha_shape import alpha_shape
                alpha = alpha_shape(points, bound_ratio=0.01) #0.0025) #0.01) #0.03)#0.01)
                if len(alpha) > 0:
                    outlines[tag] =  alpha
            except Exception as e: #scipy.spatial._qhull.QhullError as e:
                warnings.warn("Failed to compute alpha shape")
                import scipy.spatial
                outlines[tag] = points[scipy.spatial.ConvexHull(points).vertices]




def _get_frame_outlines(model):
    section_outlines = {}
    for name,section in model["sections"].items():
        _add_section_shape(section, model["sections"], section_outlines, model.ndm)

    # Function to check if list of lists is homogeneous
    homogeneous = lambda lst: (
            isinstance(lst, list) and \
              all(isinstance(x, list) and \
                  len(set(map(len, lst))) == 1 and all(isinstance(xx, list) and len(set(map(len, x))) == 1 for xx in x) for x in lst
              )
    )

    outlines = {}
    for elem in model["assembly"].values():
        elem_shapes = []
        if "sections" in elem:
            elem_shapes = [
                SectionGeometry(section_outlines[int(i)]) for i in elem["sections"]
                if i in section_outlines and section_outlines[i] is not None
            ]
        elif "section" in elem: # Truss
            if int(elem["section"]) not in section_outlines:
                continue
            elem_shapes = [
                SectionGeometry(section_outlines[int(elem["section"])])
            ]*2

        if len(elem_shapes) != 0:
            outlines[elem["name"]] = elem_shapes

            # if not homogeneous(elem_shapes):
            #     elem_shapes = np.array(elem_shapes[0])
            # else:
            #     elem_shapes = np.array(elem_shapes)

            # outlines[elem["name"]] = [SectionGeometry(shape) for shape in elem_shapes]


    return outlines

